# Copyright (c) 2015-2017  Barnstormer Softworks, Ltd.

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from __future__ import absolute_import

import xmlrpclib

from ..constants import SLICE_ROLE, PROJECT_ROLE, REQCTX, REQSTATUS
from .util import _rpcpost

DATE_FMT = "%Y-%m-%dT%H:%M:%SZ"

# pylint: disable=unsubscriptable-object
def _lookup (url, root_bundle, cert, key, typ, cred_strings, options):
  req_data = xmlrpclib.dumps((typ, cred_strings, options), methodname="lookup")
  return _rpcpost(url, req_data, (cert, key), root_bundle)

def get_version (url, root_bundle, cert, key, options = None):
  if not options: options = {}
  req_data = xmlrpclib.dumps(tuple(), methodname = "get_version")
  return _rpcpost(url, req_data, (cert, key), root_bundle)

def lookup_key_info (url, root_bundle, cert, key, cred_strings, user_urn):
  options = {"match" : {"KEY_MEMBER" : user_urn} }
  return _lookup(url, root_bundle, cert, key, "KEY", cred_strings, options)

def lookup_service_info (url, root_bundle, cert, key, cred_strings, service_type):
  options = {"match" : {"SERVICE_TYPE" : service_type} }
  return _lookup(url, root_bundle, cert, key, "SERVICE", cred_strings, options)

def lookup_member_info (url, root_bundle, cert, key, cred_strings, urn = None, uid = None,
                        email = None, lastname = None):
  match = {}
  if urn: match["MEMBER_URN"] = urn
  if uid: match["MEMBER_UID"] = uid
  if email: match["MEMBER_EMAIL"] = email
  if lastname: match["MEMBER_LASTNAME"] = lastname
  options = {"match" : match}

  return _lookup(url, root_bundle, cert, key, "MEMBER", cred_strings, options)

def create_key_info (url, root_bundle, cert, key, cred_strings, data):
  req_data = xmlrpclib.dumps(("KEY", cred_strings, {"fields" : data}), methodname="create")
  return _rpcpost(url, req_data, (cert, key), root_bundle)

def get_credentials (url, root_bundle, cert, key, creds, target_urn):
  req_data = xmlrpclib.dumps((target_urn, creds, {}), methodname="get_credentials")
  return _rpcpost(url, req_data, (cert, key), root_bundle)

def create_slice (url, root_bundle, cert, key, cred_strings, name, proj_urn, exp = None, desc = None):
  fields = {}
  fields["SLICE_NAME"] = name
  if proj_urn: fields["SLICE_PROJECT_URN"] = proj_urn
  if exp: fields["SLICE_EXPIRATION"] = exp.strftime(DATE_FMT)
  if desc: fields["SLICE_DESCRIPTION"] = desc

  req_data = xmlrpclib.dumps(("SLICE", cred_strings, {"fields" : fields}), methodname = "create")
  return _rpcpost(url, req_data, (cert, key), root_bundle)

def update_slice (url, root_bundle, cert, key, cred_strings, slice_urn, fields):
  req_data = xmlrpclib.dumps(("SLICE", slice_urn, cred_strings, {"fields" : fields}), methodname = "update")
  return _rpcpost(url, req_data, (cert, key), root_bundle)

def lookup_slices_for_member (url, root_bundle, cert, key, cred_strings, member_urn):
  options = {}
  req_data = xmlrpclib.dumps(("SLICE", member_urn, cred_strings, options), methodname = "lookup_for_member")
  return _rpcpost(url, req_data, (cert, key), root_bundle)

def lookup_slices_for_project (url, root_bundle, cert, key, cred_strings, project_urn):
  options = {"match" : {"SLICE_PROJECT_URN" : project_urn} }
  return _lookup(url, root_bundle, cert, key, "SLICE", cred_strings, options)

def lookup_slice_members (url, root_bundle, cert, key, cred_strings, slice_urn):
  options = {}
  req_data = xmlrpclib.dumps(("SLICE", slice_urn, cred_strings, options), methodname = "lookup_members")
  return _rpcpost(url, req_data, (cert, key), root_bundle)

def create_project (url, root_bundle, cert, key, cred_strings, name, exp, desc = None):
  fields = {}
  fields["PROJECT_EXPIRATION"] = exp.strftime(DATE_FMT)
  fields["PROJECT_NAME"] = name
  if desc is not None:
    fields["PROJECT_DESCRIPTION"] = desc

  req_data = xmlrpclib.dumps(("PROJECT", cred_strings, {"fields" : fields}), methodname = "create")
  return _rpcpost(url, req_data, (cert, key), root_bundle)

def delete_project (url, root_bundle, cert, key, cred_strings, project_urn):
  """Delete project by URN
  .. note::
    You may or may not be able to delete projects as a matter of policy for the given authority."""

  options = {}

  req_data = xmlrpclib.dumps(("PROJECT", project_urn, cred_strings, options), methodname = "delete")
  return _rpcpost(url, req_data, (cert, key), root_bundle)

#def _update_project (url, root_bundle, cert, key, cred_strings, project_urn):
#  options = {"fields" : {}}
#  req_data = xmlrpclib.dumps(("PROJECT", project_urn, cred_strings, options), methodname = "update")
#  s = requests.Session()
#  s.mount(url, GCU.TLSHttpAdapter())
#  resp = s.post(url, req_data, cert=(cert,key), verify=root_bundle, headers = headers())
#  return xmlrpclib.loads(resp.content)[0][0]


def lookup_projects (url, root_bundle, cert, key, cred_strings, urn = None, uid = None, expired = None):
  options = { }
  match = { }
  if urn is not None:
    match["PROJECT_URN"] = urn
  if uid is not None:
    match["PROJECT_UID"] = uid
  if expired is not None:
    match["PROJECT_EXPIRED"] = expired

  if match:
    options["match"] = match

  return _lookup(url, root_bundle, cert, key, "PROJECT", cred_strings, options)

def lookup_projects_for_member (url, root_bundle, cert, key, cred_strings, member_urn, expired = None):
  options = {}
  match = {}

  if expired is not None:
    match["PROJECT_EXPIRED"] = expired

  if match:
    options["match"] = match

  req_data = xmlrpclib.dumps(("PROJECT", member_urn, cred_strings, options), methodname = "lookup_for_member")
  return _rpcpost(url, req_data, (cert, key), root_bundle)

def lookup_project_members (url, root_bundle, cert, key, cred_strings, project_urn):
  options = {}

  req_data = xmlrpclib.dumps(("PROJECT", project_urn, cred_strings, options), methodname = "lookup_members")
  return _rpcpost(url, req_data, (cert, key), root_bundle)

def lookup_aggregates (url, root_bundle, cert, key):
  options = {"match" : {'SERVICE_TYPE': 'AGGREGATE_MANAGER'}}

  return _lookup(url, root_bundle, cert, key, "SERVICE", [], options)


def modify_slice_membership (url, root_bundle, cert, key, cred_strings, slice_urn, add = None, remove = None, change = None):
  options = {}
  if add:
    to_add = []
    for urn,role in add:
      to_add.append({"SLICE_MEMBER" : urn, "SLICE_ROLE" : role})
    options["members_to_add"] = to_add
  if remove:
    options["members_to_remove"] = remove
  if change:
    to_change = []
    for urn,role in change:
      to_change.append({"SLICE_MEMBER" : urn, "SLICE_ROLE" : role})
    options["members_to_change"] = to_change

  req_data = xmlrpclib.dumps(("SLICE", slice_urn, cred_strings, options), methodname = "modify_membership")
  return _rpcpost(url, req_data, (cert, key), root_bundle)

def modify_project_membership (url, root_bundle, cert, key, cred_strings, project_urn, add = None, remove = None, change = None):
  options = {}
  if add:
    to_add = []
    for urn,role in add:
      to_add.append({"PROJECT_MEMBER" : urn, "PROJECT_ROLE" : role})
    options["members_to_add"] = to_add
  if remove:
    options["members_to_remove"] = remove
  if change:
    to_change = []
    for urn,role in change:
      to_change.append({"PROJECT_MEMBER" : urn, "PROJECT_ROLE" : role})
    options["members_to_change"] = to_change

  req_data = xmlrpclib.dumps(("PROJECT", project_urn, cred_strings, options), methodname = "modify_membership")
  return _rpcpost(url, req_data, (cert, key), root_bundle)

def get_pending_requests (url, root_bundle, cert, key, cred_strings, member_uid, project_uid):
  req_data = xmlrpclib.dumps((member_uid, REQCTX.PROJECT, project_uid, cred_strings, {}),
                             methodname="get_pending_requests_for_user")
  return _rpcpost(url, req_data, (cert, key), root_bundle)

def resolve_request (url, root_bundle, cert, key, cred_strings, request_id, resolution, desc):
  req_data = xmlrpclib.dumps((REQCTX.PROJECT, request_id, resolution, desc, cred_strings, {}),
                             methodname="resolve_pending_request")
  return _rpcpost(url, req_data, (cert, key), root_bundle)

def create_request (url, root_bundle, cert, key, cred_strings, project_id, desc):
  JOIN = 0
  DUMMY_ATTRS = ""
  req_data = xmlrpclib.dumps((REQCTX.PROJECT, project_id, JOIN, desc, DUMMY_ATTRS, cred_strings, {}),
                             methodname="create_request")
  return _rpcpost(url, req_data, (cert,key), root_bundle)




# Copyright (c) 2014-2017 The University of Utah and Barnstormer Softworks, Ltd.

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""Library for dealing with scripts that are run in the context of a portal."""

from __future__ import absolute_import

import sys
import os
import atexit
import warnings
import json
import argparse
from argparse import Namespace

from .rspec import igext
from .rspec import pgmanifest
from .rspec.pg import Request

class ParameterType (object):
  """Parameter types understood by Context.defineParameter()."""

  INTEGER     = "integer"       #: Simple integer
  STRING      = "string"        #: Any string
  BOOLEAN     = "boolean"       #: True/False
  IMAGE       = "image"         #: URN specifying a particular image
  AGGREGATE   = "aggregate"     #: URN specifying an Aggregate Manger
  NODETYPE    = "nodetype"      #: String specifying a type of node
  BANDWIDTH   = "bandwidth"     #: Floating-point number to be used for bandwidth
  LATENCY     = "latency"       #: Floating-point number to be used for latency
  SIZE        = "size"          #: Integer for size (eg. MB, GB, etc.)
  PUBKEY      = "pubkey"        #: An RSA public key.
  LOSSRATE    = "lossrate"      #: Floating-point number 0.0 <= N < 1.0

  argparsemap = { INTEGER: int, STRING: str, BOOLEAN: bool, IMAGE: str,
                  AGGREGATE: str, NODETYPE: str, BANDWIDTH: float,
                  LATENCY: float, SIZE: int, PUBKEY: str, LOSSRATE: float }

class Parameter(object):
  UNSET = (-1, -2, -1)

  def __init__ (self):
    self.name = None
    self.description = None
    self.type = None
    self.defaultValue = None
    self.legalValues = None
    self.longDescription = None
    self.advanced = False
    self.hide = False
    self.prefix = "emulab.net.parameter."
    self.groupId = None

    self.is_set = False

  def __contains__ (self, key):
    return self.__dict__.has_key(key)

  def __getitem__ (self, key):
    return self.__dict__[key]

  def __setitem__ (self, key, val):
    self.__dict__[key] = val


class Context (object):
  """Handle context for scripts being run inside a portal.

  This class handles context for the portal, including where to put output
  RSpecs and handling parameterized scripts.

  Scripts using this class can also be run "standalone" (ie. not by the
  portal), in which case they take parameters on the command line and put
  RSpecs on the standard output.

  This class is a singleton. Most programs should access it through the
  portal.context variable; any additional "instances" of the object will
  be references to this."""

  """This is a singleton class; only one can exist at a time

  This is implemented by overriding __new__"""
  _instance = None
  _initialized = False
  def __new__(cls, *args, **kwargs):
    if not cls._instance:
      cls._instance = super(Context, cls).__new__(cls, *args, **kwargs)
    return cls._instance

  def __init__ (self):
    # If someone accidentally calls the constructor on the singleton, this
    # prevents us for wiping out its previous state
    if self.__class__._initialized:
      return

    self._request = None
    self._suppressAutoPrint = False
    self._parameters = {}
    self._parameterGroups = {}
    self._parameterOrder = []
    self._parameterErrors = []
    self._parameterWarnings = []
    self._parameterWarningsAreFatal = False
    self._bindingDone = False
    if 'GENILIB_PORTAL_MODE' in os.environ:
      self._standalone = False
      self._portalRequestPath = os.environ.get('GENILIB_PORTAL_REQUEST_PATH',None)
      self._dumpParamsPath = os.environ.get('GENILIB_PORTAL_DUMPPARAMS_PATH',None)
      self._readParamsPath = os.environ.get('GENILIB_PORTAL_PARAMS_PATH',None)
      self._parameterWarningsAreFatal = \
        bool(os.environ.get('GENILIB_PORTAL_WARNINGS_ARE_FATAL',None))
    else:
      self._standalone = True
      self._portalRequestPath = None
    self.__class__._initialized = True

  def bindRequestRSpec (self, rspec):
    """Bind the given request RSpec to the context, so that it can be
    automatically used with methods like printRequestRSpec.

    At the present time, only one request can be bound to a context"""
    if self._request is None:
      self._request = rspec
      # This feature removed until we can think through all the corner cases
      # better
      #sys.excepthook = self._make_excepthook()
      #atexit.register(self._autoPrintRequest)
    else:
      raise MultipleRSpecError

  def makeRequestRSpec (self):
    """Make a new request RSpec, bind it to this context, and return it"""
    rspec = Request()
    self.bindRequestRSpec(rspec)
    return rspec

  def printRequestRSpec (self, rspec = None):
    """Print the given request RSpec, or the one bound to this context if none
    is given.

    If run standalone (not in the portal), the request will be printed to the
    standard output; if run in the portal, it will be placed someplace the
    portal can pick it up.

    If the given rspec does not have a Tour object, this will attempt to
    build one from the file's docstring"""
    if rspec is None:
      if self._request is not None:
        rspec = self._request
      else:
        raise NoRSpecError("None supplied or bound to context")

    if not rspec.hasTour():
      tour = igext.Tour()
      if tour.useDocstring():
        rspec.addTour(tour)

    if any(self._parameters):
      rspec.ParameterData(self._parameters)

    self._suppressAutoPrint = True

    rspec.writeXML(self._portalRequestPath)

  def defineParameter (self, name, description, typ, defaultValue, legalValues = None,
                       longDescription = None, advanced = False, groupId = None, hide=False,
                       prefix="emulab.net.parameter."):
    """Define a new paramter to the script.

    The given name will be used when parameters are bound. The description is
    brief help text that will be shown to the user when making his/her selection. The
    type should be one of the types defined by ParameterType. defaultValue is
    required, but legalValues (a list) is optional; the defaultValue must be
    one of the legalValues. Entries in the legalValues list may be either
    simple strings (eg. "m400"), in which case they will be show directly to
    the user, or 2-element tuples (eg. ("m400", "ARM64"),), in which the second
    entry is what is shown to the user. defaultValue may be a tuple, so that
    one can pass, say, 'legalvalues[0]' for the option. The longDescription is
    an optional, detailed description of this parameter and how it relates to
    other parameters; it will be shown to the user if they ask to see the help,
    or as a pop-up/tooltip.  advanced, group, and groupName all provide parameter
    group abstractions.  Parameter groups are hidden by default from the user,
    and the user can expand them to view and modify them if desired.  By setting
    advanced to True, you create a parameter group named "Advanced Parameters";
    this group will not exist or be shown if none of your parameters set the
    'advanced' argument to True.

    After defining parameters, bindParameters() must be called exactly once."""

    if isinstance(defaultValue, tuple):
      defaultValue = defaultValue[0]

    if legalValues and defaultValue not in Context._legalList(legalValues):
      raise IllegalParameterDefaultError(defaultValue)

    self._parameterOrder.append(name)

    param = Parameter()
    param.name = name
    param.description = description
    param.type = typ
    param.defaultValue = defaultValue
    param.legalValues = legalValues
    param.longDescription = longDescription
    param.advanced = advanced
    param.hide = hide
    param.prefix = prefix

    if groupId is not None:
      param.groupId = groupId

    self._parameters[name] = param

    if len(self._parameters) == 1:
      atexit.register(self._checkBind)

  def defineParameterGroup(self, groupId, groupName):
    """
    Define a parameter group.  Parameters may be added to this group, which has
    an identifying token composed of alphanumeric characters (groupId), and a
    human-readable name (groupName).  Groups are intended to be used for advanced
    parameters; in the portal UI, they hidden in an expandable panel with the
    groupName --- and the user can choose to see and modify them, or not.  You
    do not need to specify any groups; you can simply stuff all your parameters
    into the "Advanced Parameters" group by setting the 'advanced' argument of
    defineParameter to True.  If you need multiple groups, define your own
    groups this way.
    """
    self._parameterGroups[groupId] = groupName

  def bindParameters (self,altParamSrc=None):
    """Returns values for the parameters defined by defineParameter().

    Returns a Namespace (like argparse), so if you call foo = bindParameters(), a
    parameter defined with name "bar" is accessed as foo.bar . Since defaults
    are required, all parameters are guaranteed to have values in the Namespace

    If run standaline (not in the portal), parameters are pulled from the command
    line (try running with --help); if run in the portal, they are pulled from
    the portal itself.  Or, if you provide the altParamSrc argument, you can
    specify your own parameters.  If altParamSrc is a dict, we will bind the
    params as a dict, using the keys as parameter names, and the values as
    parameter values.  If altParamSrc is a geni.rspec.pgmanifest.Manifest, we
    will extract the parameters and their values from the Manifest.  Finally,
    if altParamSrc is a string, we'll try to parse it as a PG manifest xml
    document.  No other forms of altParamSrc are currently specified."""
    self._bindingDone = True
    if altParamSrc:
      if isinstance(altParamSrc, dict):
        return self._bindParametersDict(altParamSrc)
      elif isinstance(altParamSrc, pgmanifest.Manifest):
        return self._bindParametersManifest(altParamSrc)
      elif isinstance(altParamSrc, (str, unicode)):
        try:
          manifestObj = pgmanifest.Manifest(xml=altParamSrc)
          return self._bindParametersManifest(manifestObj)
        except:
          ex = sys.exc_info()[0]
          raise ParameterBindError("assumed str altParamSrc was xml manifest, but"
                                   " parse error: %s" % (str(ex),))
      else:
        raise ParameterBindError("unknown altParamSrc type: %s"
                                 % (str(type(altParamSrc)),))
    elif self._standalone:
      return self._bindParametersCmdline()
    else:
      if self._dumpParamsPath:
        self._dumpParamsJSON()
      return self._bindParametersEnv()

  def makeParameterWarningsFatal (self):
    """
    Enable this option if you want to return an error to the user for
    incorrect parameter values, even if they can be autocorrected.  This can
    be useful to show the user that
    """
    self._parameterWarningsAreFatal = True

  def reportError (self,parameterError,immediate=False):
    """
    Report a parameter error to the portal.  @parameterError is an
    exception object of type ParameterError.  If @immediate is True,
    your script will exit immediately at this point with a dump of the
    errors (and fatal warnings, if enabled via
    Context.makeParameterWarningsFatal) in JSON format.  If @immediate
    is False, the errors will accumulate until Context.verifyParameters
    is called (and the errors will then be printed).
    """
    self._parameterErrors.append(parameterError)
    if immediate:
      self.verifyParameters()

  def reportWarning (self,parameterError):
    """
    Record a parameter warning.  Warnings will be printed if there are
    other errors or if warnings have been set to be fatal, when
    Context.verifyParameters() is called, or when there is another
    subsequent immediate error.
    """
    self._parameterWarnings.append(parameterError)

  def verifyParameters (self):
    """
    If there have been calls to Context.parameterError, and/or to
    Context.parameterWarning (and Context.makeParameterWarningsFatal has
    been called, making warnings fatal), this function will spit out some
    nice JSON-formatted exception info on stderr
    """
    if len(self._parameterErrors) == 0 \
      and (len(self._parameterWarnings) == 0 \
        or not self._parameterWarningsAreFatal):
      return 0

    #
    # Dump a JSON list of typed errors.
    #
    ea = []
    ea.extend(self._parameterErrors)
    ea.extend(self._parameterWarnings)
    json.dump(ea,sys.stderr,cls=PortalJSONEncoder)

    #
    # Exit with a count of errors and (fatal) warnings, added to 100 ...
    # try to distinguish ourselves meaningfully!
    #
    retcode = len(self._parameterErrors)
    if self._parameterWarningsAreFatal:
      retcode += len(self._parameterWarnings)
    sys.exit(100+retcode)

  def suppressAutoPrint (self):
    """
    Suppress the automatic printing of the bound RSpec that normally happens
    when the program exits.
    """
    self._suppressAutoPrint = True

  @staticmethod
  def _legalList(l):
    return [x if not isinstance(x, tuple) else x[0] for x in l]

  def _bindParametersCmdline (self):
    parser = argparse.ArgumentParser()
    for name in self._parameterOrder:
      opts = self._parameters[name]
      if opts['legalValues']:
        legal = Context._legalList(opts['legalValues'])
      else:
        legal = None
      if 'GENILIB_PARAMS_ASK' in os.environ:
        parser.add_argument("--" + name,
                            type    = ParameterType.argparsemap[opts['type']],
                            default = Parameter.UNSET,
                            choices = legal,
                            help    = opts['description'])
      else:
        parser.add_argument("--" + name,
                            type    = ParameterType.argparsemap[opts['type']],
                            default = opts['defaultValue'],
                            choices = legal,
                            help    = opts['description'])

    args = parser.parse_args()
    for name in self._parameterOrder:
      val = getattr(args, name)
      if val != Parameter.UNSET:
        self._parameters[name]['value'] = val
        self._parameters[name].is_set = True

    for name in self._parameterOrder:
      param = self._parameters[name]
      if param.is_set:
        continue

      while True:
        val = raw_input("%s (%s) [%s]: " % (param.description, param.type, str(param.defaultValue)))
        if val == "?":
          if param.longDescription:
            print param.longDescription
          else:
            print "No help available for this parameter"
        elif not val:
          val = param.defaultValue
          break
        else:
          break
      param.value = ParameterType.argparsemap[param.type](val)
      param.is_set = True

      setattr(args, name, param.value)

    return args

  def _bindParametersEnv (self):
    namespace = Namespace()
    paramValues = {}
    if self._readParamsPath:
      f = open(self._readParamsPath, "r")
      paramValues = json.load(f)
      f.close()
    for name in self._parameterOrder:
      opts = self._parameters[name]
      val = paramValues.get(name, opts['defaultValue'])
      try:
        val = ParameterType.argparsemap[opts['type']](val)
      except Exception:
        self.reportError(ParameterError("Could not coerce '%s' to '%s'" %
                                        (val, opts['type']), [name]))
        continue
      if opts['legalValues'] and val not in Context._legalList(opts['legalValues']):
        self.reportError(ParameterError("Illegal value '%s'" % (val,), [name]))
      else:
        setattr(namespace, name, val)
        self._parameters[name]['value'] = val
    # This might not return.
    self.verifyParameters()
    return namespace

  def _bindParametersDict(self,paramValues):
    namespace = Namespace()
    for name in self._parameterOrder:
      opts = self._parameters[name]
      val = paramValues.get(name, opts['defaultValue'])
      try:
        if type(opts['defaultValue']) == bool:
          if val == "False" or val == "false":
            val = False
          elif val == "True" or val == "true":
            val = True
          else:
            val = ParameterType.argparsemap[opts['type']](val)
        else:
          val = ParameterType.argparsemap[opts['type']](val)
      except:
        print "ERROR: Could not coerce '%s' to '%s'" % (val, opts['type'])
        continue
      if opts['legalValues'] and \
        val not in Context._legalList(opts['legalValues']):
        print "ERROR: Illegal value '%s'" % (val,),[name]
      else:
        setattr(namespace, name, val)
        self._parameters[name]['value'] = val
    # This might not return.
    self.verifyParameters()
    self._bindingDone = True
    return namespace

  def _bindParametersManifest(self,manifest):
    pdict = {}
    for manifestParameter in manifest.parameters:
      pdict[manifestParameter.name] = manifestParameter.value
    return self._bindParametersDict(pdict)

  def _dumpParamsJSON (self):
    #
    # Output the parameter dict in sorted order (sorted in terms of parameter
    # definition order).  This is correct, identical to json.dump (other than
    # key order), and much easier than subclassing json.JSONEncoder :).
    #
    didFirst = False
    f = open(self._dumpParamsPath, "w+")
    f.write('{')
    for name in self._parameterOrder:
      if didFirst:
        f.write(', ')
      else:
        didFirst = True
      opts = self._parameters[name]
      if opts.has_key('groupId') \
        and self._parameterGroups.has_key(opts['groupId']):
        opts['groupName'] = self._parameterGroups[opts['groupId']]
      json.dump(name,f)
      f.write(': ')
      json.dump(opts,f)
    f.write('}')
    f.close()
    return

  def _checkBind (self):
    if len(self._parameters) > 0 and not self._bindingDone:
      warnings.warn("Parameters were defined, but never bound with " +
                    " bindParameters()", RuntimeWarning)

  def _autoPrintRequest (self):
    if not self._suppressAutoPrint:
      self.printRequestRSpec()

  def _make_excepthook (self):
    old_excepthook = sys.excepthook
    def _excepthook(type, value, traceback):
      self.suppressAutoPrint()
      return old_excepthook(type, value, traceback)
    return _excepthook

context = Context()
"""
Module-global Context object - most users of this module should simply use
this rather than trying to create a new Context object
"""

def get_context():
  return context

class PortalJSONEncoder(json.JSONEncoder):
  def default(self, o):
    if isinstance(o,PortalError):
      return o.__objdict__()
    else:
      # First try the default encoder:
      try:
        return json.JSONEncoder.default(self, o)
      except Exception:
        try:
          # Then try to return a string, at least
          return str(o)
        except Exception:
          # Let the base class default method raise the TypeError
          return json.JSONEncoder.default(self, o)

#
# Define some exceptions.  Everybody should subclass PortalError.
#
class PortalError (Exception):
  def __init__(self,message):
    super(PortalError, self).__init__()
    self.message = message

  def __str__(self):
    return self.__class__.__name__ + ": " + self.message

  def __objdict__(self):
    retval = dict({ 'errorType': self.__class__.__name__, })
    for k in self.__dict__.keys():
      if k == 'errorType':
        continue
      if k.startswith('_'):
        continue
      retval[k] = self.__dict__[k]
    return retval


class ParameterError (PortalError):
  """
  A simple class to describe a parameter error.  If you need to report
  an error with a user-specified parameter value to the Portal UI,
  please create (don't throw) one of these error objects, and tell the
  Portal about it by calling Context.reportError.
  """
  def __init__(self,message,paramList):
    """
    Create a ParameterError.  @message is the overall error message;
    in the Portal Web UI, it will be displayed near each involved
    parameter for maximal impact.  @paramList is a list of the
    parameters that are involved in the error (often it is the
    combination of parameters that creates the error condition).
    The Portal UI will show this error message near *each* involved
    parameter to increase user understanding of the error.
    """
    super(ParameterError, self).__init__(message)
    self.params = paramList


class ParameterWarning (PortalError):
  """
  A simple class to describe a parameter warning.  If you need to
  report an warning with a user-specified parameter value to the
  Portal UI, please create (don't throw) one of these error objects,
  and tell the Portal about it by calling Context.reportWarning .  The
  first time the Portal UI runs your geni-lib script with a user's
  parameter values, it turns on the "warnings are fatal" mode (and
  then warnings are reported as errors).  This gives you a chance to
  warn the user that they might be about to do something stupid,
  and/or suggest a set of modified values that will improve the
  situation.  .
  """
  def __init__(self,message,paramList,fixedValues=None):
    """
    Create a ParameterWarning.  @message is the overall error
    message; in the Portal Web UI, it will be displayed near each
    involved parameter for maximal impact.  @paramList is a list of
    the parameters that are involved in the warning (often it is the
    combination of parameters that creates the error condition).
    The Portal UI will show this warning message near *each*
    involved parameter to increase user understanding of the error.
    If you supply the @fixedValue dict, the Portal UI will change
    the values the user submitted to those you suggest (and it will
    tell them it did so).  You might want to supply @fixedValues for
    a proper warning, because if something is only a warning, that
    implies your script can and will proceed in the absence of
    further user input.  But sometimes we want to let the user know
    that a parameter change will occur, so we warn them and
    autocorrect!
    """
    super(ParameterWarning, self).__init__(message)
    self.params = paramList
    if not fixedValues: fixedValues = {}
    self.fixedValues = fixedValues


class IllegalParameterDefaultError (PortalError):
  def __init__ (self,val):
    super(IllegalParameterDefaultError, self).__init__("no message?")
    self._val = val

  def __str__ (self):
    return "% given as a default value, but is not listed as a legal value" % self._val


class ParameterBindError (PortalError):
  def __init__ (self,val):
    super(ParameterBindError, self).__init__("no message?")
    self._val = val

  def __str__ (self):
    return "bad parameter binding: %s" % str(self._val,)


class NoRSpecError (PortalError):
  def __init__ (self,val):
    super(NoRSpecError, self).__init__("no message?")
    self._val = val

  def __str__ (self):
    return "No RSpec given: %s" % str(self._val,)

class MultipleRSpecError (PortalError):
  def __init__ (self,val):
    super(MultipleRSpecError, self).__init__("no message?")
    self._val = val

  def __str__ (self):
    return "Only one RSpec can be bound to a portal.Context"



# Copyright (c) 2014 Barnstomer Softworks, Ltd. and The University of Utah

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from __future__ import absolute_import

import inspect
import sys

from .protogeni import PGCompute

from .instageni import UtahDDC # pylint: disable=unused-import

class AptAM(PGCompute): pass

Apt = AptAM("apt", "boss.apt.emulab.net", "urn:publicid:IDN+apt.emulab.net+authority+cm")

def aggregates ():
  module = sys.modules[__name__]
  for _,obj in inspect.getmembers(module):
    if isinstance(obj, PGCompute):
      yield obj

def name_to_aggregate ():
  result = dict()
  module = sys.modules[__name__]
  for _,obj in inspect.getmembers(module):
    if isinstance(obj, PGCompute):
      result[obj.name] = obj
  return result



# Copyright (c) 2017  Barnstormer Softworks, Ltd.

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from __future__ import absolute_import

class AMTYPE(object):
  FOAM = "foam"
  VTS = "vts"
  IG = "ig"
  PG = "ig"
  EG = "eg"
  OTHER = "other"

class AMSpec(object):
  def __init__ (self):
    self.cmid = None
    self.desc = None
    self.shortname = None
    self.longname = None
    self.url = None
    self.type = None
    self.cert = None

  def __json__ (self):
    d = {"version" : 1,
         "cmid" : self.cmid,
         "desc" : self.desc,
         "longname" : self.longname,
         "shortname" : self.shortname,
         "url" : self.url,
         "type" : self.type,
         "cert" : self.cert}
    return d

  @classmethod
  def _jconstruct (cls, obj):
    ams = cls()
    ams.cmid = obj["cmid"]
    ams.desc = obj["desc"]
    ams.shortname = obj["shortname"]
    ams.longname = obj["longname"]
    ams.url = obj["url"]
    ams.type = obj["type"]
    ams.cert = obj["cert"]
    return ams

  def build (self):
    am = None
    if self.type == AMTYPE.IG:
      from .instageni import IGCompute
      am = IGCompute(self.shortname, None, self.cmid, self.url)
      am.cert_data = self.cert
    elif self.type == AMTYPE.EG:
      from .exogeni import EGCompute
      am = EGCompute(self.shortname, None, self.cmid, self.url)
      am.cert_data = self.cert
    elif self.type == AMTYPE.VTS:
      from .vts import VTS
      am = VTS(self.shortname, None, self.url)
      am.cert_data = self.cert
    if am:
      am._amspec = self
    return am


# Emulab GID PEM data often lacks begin/end markers
def fixCert (data):
  if data[0:3] != "---":
    data = "-----BEGIN CERTIFICATE-----\n%s\n-----END CERTIFICATE-----\n" % (data)
  return data




# Copyright (c) 2016-2018  Barnstormer Softworks, Ltd.

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from .exceptions import AMError

# pylint: disable=multiple-statements
class ProtoGENIError(AMError): pass

class ResourceBusyError(ProtoGENIError): pass
class VLANUnavailableError(ProtoGENIError): pass
class InsufficientBandwidthError(ProtoGENIError): pass
class InsufficientNodesError(ProtoGENIError): pass
class InsufficientMemoryError(ProtoGENIError): pass
class NoMappingError(ProtoGENIError): pass
# pylint: enable=multiple-statements

def raiseError(res):
  amcode = res["code"]["am_code"]
  output = res["output"]
  if amcode == 14:
    e = ResourceBusyError(output, res)
  elif amcode == 24:
    e = VLANUnavailableError(output, res)
  elif amcode == 25:
    e = InsufficientBandwidthError(output, res)
  elif amcode == 26:
    e = InsufficientNodesError(output, res)
  elif amcode == 27:
    e = InsufficientMemoryError(output, res)
  elif amcode == 28:
    e = NoMappingError(output, res)
  else:
    e = ProtoGENIError(output, res)

  try:
    e.error_url = res["code"]["protogeni_error_url"]
  except KeyError:
    pass

  raise e



# Copyright (c) 2014-2016  Barnstormer Softworks, Ltd.

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from __future__ import absolute_import, print_function

import json

import requests

from . import germ

class Connection(germ.Connection):
  def __init__ (self):
    super(Connection, self).__init__()
    self.user = "foamadmin"

  @property
  def pgcmid (self):
    url = "%s/core/admin/vts/pg-cmid" % (self.baseurl)
    r = requests.get(url, **self.rkwargs)
    return r.json()["value"]

  @pgcmid.setter
  def pgcmid (self, val):
    url = "https://%s:%d/core/admin/vts/pg-cmid/%s" % (self.host, self.port, val)
    requests.post(url, **self.rkwargs)

  @property
  def pgvlans (self):
    url = "https://%s:%d/core/admin/vts/pgvlans" % (self.host, self.port)
    r = requests.get(url, **self.rkwargs)
    return r.json()["value"]

  @property
  def slivers (self):
    url = "https://%s:%d/core/admin/vts/slivers" % (self.host, self.port)
    r = requests.get(url, **self.rkwargs)
    return r.json()["value"]

  @property
  def images (self):
    url = "https://%s:%d/core/admin/vts/images" % (self.host, self.port)
    r = requests.get(url, **self.rkwargs)
    return r.json()["value"]

  @property
  def bridges (self):
    url = "https://%s:%d/core/admin/vts/ovs/bridges" % (self.host, self.port)
    r = requests.get(url, **self.rkwargs)
    return r.json()["value"]

  def deleteSlivers (self, slice_urn):
    url = "%s/core/admin/vts/slice/%s" % (self.baseurl, slice_urn)
    r = requests.delete(url, **self.rkwargs)
    return r.json()["value"]

  def addTargetBridge (self, name, brname):
    url = "https://%s:%d/core/admin/vts/target-bridge" % (self.host, self.port)
    d = json.dumps({"name" : name, "brname" : brname})
    requests.post(url, d, **self.rkwargs)

  def addPGVlan (self, name, vid):
    url = "https://%s:%d/core/admin/vts/pgvlan" % (self.host, self.port)
    d = json.dumps({"name" : name, "vid" : vid})
    requests.post(url, d, **self.rkwargs)

  def setSSLVPNIP (self, ipstr):
    url = "https://%s:%d/core/admin/vts/vf/sslvpn/local-ip" % (self.host, self.port)
    d = json.dumps(ipstr)
    requests.post(url, d, **self.rkwargs)

  def addCircuitPlane (self, typ, label, endpoint, mtu, types = None, encoded = True):
    if not types: types = []

    url = "https://%s:%d/core/admin/vts/circuitplane/%s" % (self.host, self.port, typ)
    d = json.dumps({"label" : label, "endpoint" : endpoint,
                    "supported-types" : types, "encoded" : encoded, "mtu" : mtu})
    r = requests.put(url, d, **self.rkwargs)
    return r

  def removeCircuitPlane (self, label):
    url = "https://%s:%d/core/admin/vts/circuitplane/%s" % (self.host, self.port, label)
    requests.delete(url, **self.rkwargs)

  def getDatapaths (self, sliver_urn):
    url = "https://%s:%d/core/admin/vts/sliver/%s/datapaths" % (self.host, self.port, sliver_urn)
    r = requests.get(url, **self.rkwargs)
    return r.json()["value"]

  def deleteDatapath (self, dpname, really = False):
    if not really:
      print("WARNING!!!!  You do not want to do this to datapaths managed by a sliver unless you really know what you are doing.")
      return
    url = "https://%s:%d/core/admin/vts/ovs/bridge/%s" % (self.host, self.port, dpname)
    r = requests.delete(url, **self.rkwargs)
    return r.json()["value"]

  def getContainers (self, sliver_urn):
    url = "https://%s:%d/core/admin/vts/sliver/%s/containers" % (self.host, self.port, sliver_urn)
    r = requests.get(url, **self.rkwargs)
    return r.json()["value"]

  def getRequestRspec (self, sliver_urn):
    url = "https://%s:%d/core/admin/vts/sliver/%s/request-rspec" % (self.host, self.port, sliver_urn)
    r = requests.get(url, **self.rkwargs)
    return r.json()["value"]

  def getManifestRspec (self, sliver_urn):
    url = "https://%s:%d/core/admin/vts/sliver/%s/manifest-rspec" % (self.host, self.port, sliver_urn)
    r = requests.get(url, **self.rkwargs)
    return r.json()["value"]

  def removePort (self, sliver_urn, dpname, client_id):
    url = "https://%s:%d/core/admin/vts/sliver/%s/datapath/%s/port/%s" % (self.host, self.port,
                                                                          sliver_urn, dpname, client_id)
    r = requests.delete(url, **self.rkwargs)
    return r.json()["value"]

  def addPGLocal (self, sliver_urn, dpname, client_id, pgcircuit):
    url = "https://%s:%d/core/admin/vts/sliver/%s/datapath/%s/port/%s" % (self.host, self.port,
                                                                          sliver_urn, dpname, client_id)
    d = json.dumps({"type" : "pg-local", "shared-lan" : pgcircuit})
    r = requests.put(url, d, **self.rkwargs)
    return r.json()["value"]

  def addImage (self, image_name):
    url = "https://%s:%d/core/admin/vts/image" % (self.host, self.port)
    d = json.dumps({"name" : image_name})
    r = requests.post(url, d, **self.rkwargs)
    return r.json()["value"]

  def setURNLock (self, sliver_urn, user_urn):
    url = "https://%s:%d/core/admin/vts/sliver/%s/lock" % (self.host, self.port, sliver_urn)
    d = json.dumps({"user-urn" : user_urn})
    r = requests.put(url, d, **self.rkwargs)
    return r.json()["value"]

  def deleteDropboxVol (self, sliver_uuid):
    url = "https://%s:%d/core/admin/vts/dropbox/%s" % (self.host, self.port, sliver_uuid)
    r = requests.delete(url, **self.rkwargs)
    return r.json()["value"]

  def deallocateSliver (self, sliver_urn):
    url = "https://%s:%d/core/admin/vts/sliver/%s/deallocate" % (self.host, self.port, sliver_urn)
    r = requests.get(url, **self.rkwargs)
    return r.json()["value"]



# Copyright (c) 2016 The University of Utah

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from __future__ import absolute_import

from ..pg import RawPC, Request
from .pndefs import PNDEFS
import geni.urn

class eNodeB(RawPC):
    _ENODEB_OS = "emulab-ops:GENERICDEV-NOVLANS"

    def __init__ (self, client_id, component_id = None):
        super(eNodeB, self).__init__(client_id, component_id = component_id)
        self.hardware_type = "enodeb"  # set hwtype to general eNB node class.
        self.disk_image = geni.urn.Image(PNDEFS.PNET_AM, eNodeB._ENODEB_OS)

Request.EXTENSIONS.append(("eNodeB", eNodeB))



# Copyright (c) 2016 The University of Utah

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from __future__ import absolute_import

from lxml import etree as ET

from ..pg import RawPC, Namespaces, Request
from .pndefs import PNDEFS

class UE(RawPC):
    def __init__ (self, client_id, component_id = None):
        super(UE, self).__init__(client_id, component_id = component_id)
        self.hardware_type = "ue"
        self.adb_target = None

    def _write (self, root):
        nd = super(UE, self)._write(root)
        if self.adb_target:
            tgt = ET.SubElement(nd, "{%s}adb_target" % (Namespaces.EMULAB.name))
            tgt.attrib["target_id"] = self.adb_target
        return nd

Request.EXTENSIONS.append(("UE", UE))



# Copyright (c) 2014  Barnstormer Softworks, Ltd.

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import xmlrpclib

class Connection(object):
  def __init__ (self):
    self.user = "fvadmin"
    self.password = None
    self.host = "localhost"
    self.jsonport = 8081
    self.xmlrpcport = 8080

    self._xmlconn = None
    self._jsonconn = None

  @property
  def xmlconn (self):
    if self._xmlconn is None:
      self._xmlconn = xmlrpclib.ServerProxy("https://%s:%s@%s:%d/xmlrpc" % (
        self.user, self.password, self.host, self.xmlrpcport))
    return self._xmlconn

  def getVersion (self):
    s = self.xmlconn.api.ping("foo")
    ver = s.split("=")[1].split(":")[0].split("-")[1]
    return ver



# Copyright (c) 2014-2015  Barnstormer Softworks, Ltd.

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

class AbstractImplementationError(Exception):
  def __str__ (self):
    return "Called unimplemented abstract method"

class NoUserError(Exception):
  def __str__ (self):
    return "No framework user specified"

class SliceCredError(Exception):
  def __init__ (self, text):
    super(SliceCredError, self).__init__()
    self.text = text

  def __str__ (self):
    return self.text

class WrongNumberOfArgumentsError(Exception):
  def __str__ (self):
    return "Called a function with the wrong number of arguments"



# Copyright (c) 2015-2018  Barnstormer Softworks, Ltd.

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

# Streamlined implementation of xmlrpc calls to AM API v3-compliant aggregates
# Only uses python requests module, without a ton of extra SSL dependencies

from __future__ import absolute_import

import xmlrpclib

from .util import _rpcpost

# pylint: disable=unsubscriptable-object
def getversion (url, root_bundle, cert, key, options = None):
  if not options: options = {}
  req_data = xmlrpclib.dumps(options, methodname="GetVersion")
  return _rpcpost(url, req_data, (cert, key), root_bundle)

def poa (url, root_bundle, cert, key, creds, urns, action, options = None):
  if not options: options = {}
  if not isinstance(urns, list): urns = [urns]

  cred_list = []
  for cred in creds:
    cred_list.append({"geni_value" : open(cred.path, "rb").read(), "geni_type" : cred.type, "geni_version" : cred.version})

  req_data = xmlrpclib.dumps((urns, cred_list, action, options),
                             methodname="PerformOperationalAction")
  return _rpcpost(url, req_data, (cert, key), root_bundle)

def paa (url, root_bundle, cert, key, action, options = None):
  if not options: options = {}

  req_data = xmlrpclib.dumps((action, options),
                             methodname="PerformAggregateAction")
  return _rpcpost(url, req_data, (cert, key), root_bundle)

def allocate (url, root_bundle, cert, key, creds, slice_urn, rspec, options = None):
  if not options: options = {}

  cred_list = []
  for cred in creds:
    cred_list.append({"geni_value" : open(cred.path, "rb").read(), "geni_type" : cred.type,
                      "geni_version" : cred.version})

  req_data = xmlrpclib.dumps((slice_urn, cred_list, rspec, options),
                             methodname="Allocate")
  return _rpcpost(url, req_data, (cert, key), root_bundle)

def provision (url, root_bundle, cert, key, creds, urns, options = None):
  if not options: options = {}
  if not isinstance(urns, list): urns = [urns]

  cred_list = []
  for cred in creds:
    cred_list.append({"geni_value" : open(cred.path, "rb").read(), "geni_type" : cred.type,
                      "geni_version" : cred.version})

  req_data = xmlrpclib.dumps((urns, cred_list, options), methodname="Provision")
  return _rpcpost(url, req_data, (cert, key), root_bundle)

def delete (url, root_bundle, cert, key, creds, urns, options = None):
  if not options: options = {}
  if not isinstance(urns, list): urns = [urns]

  cred_list = []
  for cred in creds:
    cred_list.append({"geni_value" : open(cred.path, "rb").read(), "geni_type" : cred.type,
                      "geni_version" : cred.version})

  req_data = xmlrpclib.dumps((urns, cred_list, options), methodname="Delete")
  return _rpcpost(url, req_data, (cert, key), root_bundle)




 # Copyright (c) 2014-2016  Barnstormer Softworks, Ltd.

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

class User(object):
  def __init__ (self):
    self.name = None
    self.urn = None
    self._keys = []

  def addKey (self, path):
    self._keys.append(path)





# Copyright (c) 2013  Barnstormer Softworks, Ltd

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

class XMLContext(object):
  def __init__ (self, rspec, root, cur_elem = None):
    self.rspec = rspec
    self.root = root
    self.curelem = cur_elem



# Copyright (c) 2016  Barnstormer Softworks, Ltd.

#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at http://mozilla.org/MPL/2.0/.

class AMError(Exception):
  def __init__ (self, text, data = None):
    super(AMError, self).__init__()
    self.text = str(text)
    self.data = data
  def __str__ (self):
    return self.text




# Copyright (c) 2014-2017  Barnstormer Softworks, Ltd.

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from __future__ import absolute_import

import os
import os.path

class _Registry(object):
  def __init__ (self):
    self._data = {}

  def register (self, name, obj):
    self._data[name] = obj

  def get (self, name):
    return self._data[name]

def convertCH2AggregateSpecs(ch2info, path = None):
  from .spec import AMSpec, AMTYPE

  typemap = {
    "ui_instageni_am" : AMTYPE.IG,
    "ui_exogeni_am" : AMTYPE.EG,
    "ui_vts_am" : AMTYPE.VTS,
    "ui_foam_am" : AMTYPE.FOAM,
    "ui_other_am" : AMTYPE.OTHER
  }

  speclist = []
  for info in ch2info:
    ams = AMSpec()
    ams.cmid = info["SERVICE_URN"]
    ams.desc = info["SERVICE_DESCRIPTION"]
    ams.longname = info["SERVICE_NAME"]
    ams.shortname = info["_GENI_SERVICE_SHORT_NAME"]
    ams.url = info["SERVICE_URL"]
    ams.type = typemap[info["_GENI_SERVICE_ATTRIBUTES"]["UI_AM_TYPE"]]
    speclist.append(ams)

  return speclist

def loadFromRegistry (context):
  from . import frameworks
  from . import spec
  from .. import urn
  from .spec import AMSpec, AMTYPE, fixCert

  ammap = {}
  cf = context.cf
  if isinstance(cf, frameworks.CHAPI2):
    if cf.name == "emulab-ch2":
      from .frameworks import ProtoGENI
      # Make a synthetic PG CF to get the aggregates
      cf = ProtoGENI()
      cf.key = context.cf.key
      cf.cert = context.cf.cert
    else:
      svclist = convertCH2AggregateSpecs(cf.loadAggregates())
      for ams in svclist:
        am = ams.build()
        if am:
          ammap[am.name] = am
  if isinstance(cf, frameworks.ProtoGENI):
    components = cf.loadComponents(context)
    for info in components:
      u = urn.GENI(info["urn"])
      if u.name not in ["cm", "am"]:
        continue
      ams = AMSpec()
      ams.cmid = info["urn"]
      if info["hrn"][-3:] == ".cm":
        ams.shortname = info["hrn"][:-3]
        ams.url = "%s/am/2.0" % (info["url"][:-3])
      else:
        ams.shortname = info["hrn"]
        ams.url = info["url"]

      ams.cert = fixCert(info["gid"])
      if info["url"].count("instageni"):
        ams.type = AMTYPE.IG
      elif info["url"].count("protogeni"):
        ams.type = AMTYPE.PG
      elif info["url"].count("orca"):
        ams.type = AMTYPE.EG
      am = ams.build()
      if am:
        ammap[am.name] = am

  return ammap

class AM(object):
  """Base class wrapping GENI AM APIv2 and AM APIv3 functionality."""

  class UnspecifiedComponentManagerError(Exception):
    def __str__ (self):
      return "AM object does not have a component manager ID specified"

  class InvalidRSpecPathError(Exception):
    def __init__ (self, path):
      super(AM.InvalidRSpecPathError, self).__init__()
      if len(path) > 400:
        path = path[:400] + "..."
      self.path = path
    def __str__ (self):
      return "RSpec object provided as path string, but path not found: %s" % (self.path)


  def __init__ (self, name, url, api, amtype, cmid=None):
    self.url = url
    self.name = name
    self.cert_data = None
    self._cmid = cmid
    self._apistr = api
    self._api = None
    self._typestr = amtype
    self._type = None
    self._amspec = None

  @property
  def component_manager_id (self):
    if self._cmid:
      return self._cmid
    raise AM.UnspecifiedComponentManagerError()

  @property
  def api (self):
    if not self._api:
      self._api = APIRegistry.get(self._apistr)
    return self._api

  @property
  def amtype (self):
    if not self._type:
      self._type = AMTypeRegistry.get(self._typestr)
    return self._type

  def listresources (self, context, sname = None, available = False):
    """GENI AM APIv2 method to get available resources from an aggregate, or resources allocated to
    a specific sliver.

    Args:
      context: geni-lib context
      sname (str): Slice name (optional)
      available (bool): Only list available resources

    Returns:
      geni.rspec.RSpec:
        If `sname` is provided, `listresources` will return a manifest rspec for the given slice name.  Otherwise,
        `listresources` will return the advertisement rspec for the given aggregate.
    """

    rspec_data = self.api.listresources(context, self.url, sname, {"geni_available" : available})
    if sname is None:
      return self.amtype.parseAdvertisement(rspec_data)
    else:
      return self.amtype.parseManifest(rspec_data)

  def sliverstatus (self, context, sname):
    """GENI AM APIv2 method to get the status of a current sliver at the given aggregate.

    Args:
      context: geni-lib context
      sname (str): Slice name

    Returns:
      dict:
        Mapping of key/value pairs for status information the aggregate supports.
    """

    status = self.api.sliverstatus(context, self.url, sname)
    return status

  def renewsliver (self, context, sname, date):
    """GENI AM APIv2 method to renew a sliver until the given datetime.

    Args:
      context: geni-lib context
      sname (str): Slice name
      date (str): RFC 3339-compliant date string for new expiration date

    .. note::
      Aggregates may have maximum expiration limits, restricting how far in
      the future you can set your expiration.  This call may result in an
      error in such cases, or success with a sooner future date.
    """

    res = self.api.renewsliver(context, self.url, sname, date)
    return res

  def deletesliver (self, context, sname):
    """GENI AM APIv2 method to delete a resource reservation at this aggregate.

    Args:
      context: geni-lib context
      sname (str): Slice name
    """
    self.api.deletesliver(context, self.url, sname)

  def createsliver (self, context, sname, rspec):
    """GENI AM APIv2 method to reserve resources at this aggregate.

    Args:
      context: geni-lib context
      sname (str): Slice name
      rspec (geni.rspec.RSpec): Valid request RSpec
    """
    if isinstance(rspec, (str, unicode)):
      rspec = os.path.normpath(os.path.expanduser(rspec))
      if not os.path.exists(rspec):
        raise AM.InvalidRSpecPathError(rspec)
      rspec_data = open(rspec, "rb").read()
    else:
      rspec_data = rspec.toXMLString()
    res = self.api.createsliver(context, self.url, sname, rspec_data)
    return self.amtype.parseManifest(res)

  def getversion (self, context):
    """GENI AM API method to get the version information for this aggregate.

    Args:
      context: geni-lib context

    Returns:
      dict: Dictionary of key/value pairs with version information from this aggregate.
    """

    return self.api.getversion(context, self.url)


APIRegistry = _Registry()
AMTypeRegistry = _Registry()
FrameworkRegistry = _Registry()



# Copyright (c) 2013  Barnstormer Softworks, Ltd.

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from __future__ import absolute_import

from lxml import etree as ET
import geni.namespaces as GNS

class RSpec (object):
  def __init__ (self, rtype):
    self.NSMAP = {}
    self._nsnames = set()
    self._loclist = []
    self.addNamespace(GNS.XSNS)
    self.type = rtype

  def addNamespace (self, ns, prefix = ""):
    if ns.name in self._nsnames:
      return

    self._nsnames.add(ns.name)

    if prefix != "":
      self.NSMAP[prefix] = ns.name
    else:
      self.NSMAP[ns.prefix] = ns.name

    if ns.location is not None:
      self._loclist.append(ns.name)
      self._loclist.append(ns.location)

  def toXMLString (self, pretty_print = False):
    rspec = self.getDOM()
    return ET.tostring(rspec, pretty_print = pretty_print)

  def getDOM (self):
    rspec = ET.Element("rspec", nsmap = self.NSMAP)
    rspec.attrib["{%s}schemaLocation" % (GNS.XSNS.name)] = " ".join(self._loclist)
    rspec.attrib["type"] = self.type
    return rspec



# Copyright (c) 2014-2017  Barnstormer Softworks, Ltd.

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from __future__ import absolute_import

import base64
import functools
import decimal

import ipaddress
from lxml import etree as ET

import geni.rspec
import geni.namespaces as GNS
from geni.rspec.pg import Resource


class Namespaces(object):
  VTS = GNS.Namespace("vts", "http://geni.bssoftworks.com/rspec/ext/vts/request/1")
  SDN = GNS.Namespace("sdn", "http://geni.bssoftworks.com/rspec/ext/sdn/request/1")

class BadImageTypeError(Exception):
  def __init__ (self, rtype):
    self.rtype = rtype
  def __str__ (self):
    return "Supplied image must be of type %s" % (self.rtype)

################################################
# Base Request - Must be at top for EXTENSIONS #
################################################

class Request(geni.rspec.RSpec):
  EXTENSIONS = []

  def __init__ (self):
    super(Request, self).__init__("request")
    self._resources = []
    self.topo_name = None

    self.addNamespace(GNS.REQUEST, None)
    self.addNamespace(Namespaces.VTS)
    self.addNamespace(Namespaces.SDN)

    self._ext_children = []
    for name,ext in Request.EXTENSIONS:
      self._wrapext(name,ext)

  def _wrapext (self, name, klass):
    @functools.wraps(klass.__init__)
    def wrap(*args, **kw):
      instance = klass(*args, **kw)
      self._ext_children.append(instance)
      return instance
    setattr(self, name, wrap)

  def addResource (self, rsrc):
    for ns in rsrc.namespaces:
      self.addNamespace(ns)
    self._resources.append(rsrc)

  def writeXML (self, path):
    f = open(path, "w+")
    f.write(self.toXMLString(True))
    f.close()

  def toXMLString (self, pretty_print = False):
    rspec = self.getDOM()

    if self.topo_name:
      ci = ET.SubElement(rspec, "{%s}client-info" % (Namespaces.VTS))
      ci.attrib["topo-name"] = str(self.topo_name)

    for resource in self._resources:
      resource._write(rspec)

    for obj in self._ext_children:
      obj._write(rspec)

    buf = ET.tostring(rspec, pretty_print = pretty_print)
    return buf

  @property
  def resources(self):
      return self._resources + self._ext_children

######################
# Internal Functions #
######################

def _am_encrypt (gv, plaintext):
  # This should be turned into an Encryptor object that you can init once and carry around
  from cryptography.hazmat.backends import default_backend
  from cryptography.hazmat.primitives import hashes, serialization
  from cryptography.hazmat.primitives.asymmetric import padding

  pubkey = serialization.load_pem_public_key(gv["request.pubkey"], backend=default_backend())

  if gv["request.hash"] == "sha1":
    hfunc = hashes.SHA1
  elif gv["request.hash"] == "sha256":
    hfunc = hashes.SHA256
  elif gv["request.hash"] == "sha384":
    hfunc = hashes.SHA384
  elif gv["request.hash"] == "sha512":
    hfunc = hashes.SHA512

  return base64.b64encode(pubkey.encrypt(plaintext, padding.OAEP(padding.MGF1(hfunc()), hfunc(), None)))


###################
# Utility Objects #
###################

class DelayInfo(object):
  def __init__ (self, time = None, jitter = None, correlation = None, distribution = None):
    self.time = time
    self.jitter = jitter
    self.correlation = correlation
    self.distribution = distribution

  def __json__ (self):
    d = {"type" : "egress-delay"}
    if self.time: d["time"] = self.time
    if self.jitter: d["jitter"] = self.jitter
    if self.correlation: d["correlation"] = self.correlation
    if self.distribution: d["distribution"] = self.distribution
    return d

  def _write (self, element):
    d = ET.SubElement(element, "{%s}egress-delay" % (Namespaces.VTS.name))
    if self.time: d.attrib["time"] = str(self.time)
    if self.jitter: d.attrib["jitter"] = str(self.jitter)
    if self.correlation: d.attrib["correlation"] = str(self.correlation)
    if self.distribution: d.attrib["distribution"] = self.distribution
    return d


class LossInfo(object):
  def __init__ (self, percent):
    self.percent = percent

  def __json__ (self):
    return {"type" : "egress-loss", "percent" : "%d" % (self.percent)}

  def _write (self, element):
    d = ET.SubElement(element, "{%s}egress-loss" % (Namespaces.VTS))
    d.attrib["percent"] = "%d" % (self.percent)
    return d


class ReorderInfo(object):
  def __init__ (self, percent, correlation, gap = None):
    self.percent = percent
    self.correlation = correlation
    self.gap = gap

  def _write (self, element):
    d = ET.SubElement(element, "{%s}egress-reorder" % (Namespaces.VTS))
    d.attrib["percent"] = str(self.percent)
    d.attrib["correlation"] = str(self.correlation)
    if self.gap:
      d.attrib["gap"] = str(self.gap)
    return d


###################
# Datapath Images #
###################

class Image(object):
  def __init__ (self, name):
    self.name = name
    self._features = []
    self._image_attrs = []

  def setImageAttribute (self, name, val):
    self._image_attrs.append((name, val))

  def _write (self, element):
    i = ET.SubElement(element, "{%s}image" % (Namespaces.VTS.name))
    i.attrib["name"] = self.name
    for feature in self._features:
      feature._write(i)
    for (name,val) in self._image_attrs:
      ae = ET.SubElement(i, "{%s}image-attribute" % (Namespaces.VTS))
      ae.attrib["name"] = name
      ae.attrib["value"] = str(val)
    return i

class SimpleDHCPImage(Image):
  def __init__ (self, subnet = None):
    super(SimpleDHCPImage, self).__init__("uh.simple-dhcpd")
    self.subnet = subnet

  def _write (self, element):
    e = super(SimpleDHCPImage, self)._write(element)
    if self.subnet:
      subnet = ET.SubElement(e, "{%s}image-attribute" % (Namespaces.VTS))
      subnet.attrib["name"] = "subnet"
      subnet.attrib["value"] = str(self.subnet)
    return e


class DatapathImage(Image):
  pass

class OVSImage(DatapathImage):
  def __init__ (self, name):
    super(OVSImage, self).__init__(name)

  @property
  def sflow (self):
    return None

  @sflow.setter
  def sflow (self, val):
    if isinstance(val, SFlow):
      self._features.append(val)
    # TODO: Throw exception

  @property
  def netflow (self):
    return None

  @netflow.setter
  def netflow (self, val):
    if isinstance(val, NetFlow):
      self._features.append(val)
    # TODO: Throw exception

  def setMirror (self, port):
    self._features.append(MirrorPort(port))


class OVSOpenFlowImage(OVSImage):
  def __init__ (self, controller, ofver = "1.0", dpid = None):
    super(OVSOpenFlowImage, self).__init__("bss:ovs-201-of")
    self.dpid = dpid
    self.controller = controller
    self.ofver = ofver

  def _write (self, element):
    i = super(OVSOpenFlowImage, self)._write(element)
    c = ET.SubElement(i, "{%s}controller" % (Namespaces.SDN.name))
    c.attrib["url"] = self.controller

    v = ET.SubElement(i, "{%s}openflow-version" % (Namespaces.VTS.name))
    v.attrib["value"] = self.ofver

    if self.dpid:
      d = ET.SubElement(i, "{%s}openflow-dpid" % (Namespaces.VTS.name))
      d.attrib["value"] = str(self.dpid)

    return i

class UnknownSTPModeError(Exception):
  def __init__ (self, val):
    self._val = val
  def __str__ (self):
    return "Unknown STP Mode (%d)" % (self._val)

class IllegalModeForParamError(Exception):
  def __init__ (self, param):
    self.param = param
  def __str__ (self):
    return "The parameter '%s' is not configurable in the current STP mode" % (self.param)

class OVSL2STP(object):
  STP = 1
  RSTP = 2

  def __init__ (self):
    self._mode = OVSL2STP.STP
    self._rstp_params = {}
    self._stp_params = {}

  @property
  def mode (self):
    return self._mode

  @mode.setter
  def mode (self, val):
    if val != OVSL2STP.STP and val != OVSL2STP.RSTP:
      raise UnknownSTPModeError(val)
    self._mode = val

  @property
  def type (self):
    if self._mode == OVSL2STP.STP:
      return "stp"
    elif self._mode == OVSL2STP.RSTP:
      return "rstp"

  @property
  def priority (self):
    try:
      return self._stp_params["priority"]
    except KeyError:
      return None

  @priority.setter
  def priority (self, val):
    self._stp_params["priority"] = val
    self._rstp_params["priority"] = val

  @property
  def max_age (self):
    try:
      return self._stp_params["max-age"]
    except KeyError:
      return None
    
  @max_age.setter
  def max_age (self, val):
    self._stp_params["max-age"] = val
    self._rstp_params["max-age"] = val

  @property
  def hello_time (self):
    if self._mode != OVSL2STP.STP:
      raise IllegalModeForParamError("hello-time")
    try:
      return self._stp_params["hello-time"]
    except KeyError:
      return None

  @hello_time.setter
  def hello_time (self, val):
    if self._mode != OVSL2STP.STP:
      raise IllegalModeForParamError("hello-time")
    self._stp_params["hello-time"] = val

  @property
  def forward_delay (self):
    try:
      return self._stp_params["forward-delay"]
    except KeyError:
      return None
    
  @forward_delay.setter
  def forward_delay (self, val):
    self._stp_params["forward-delay"] = val
    self._rstp_params["forward-delay"] = val

  @property
  def address (self):
    try:
      return self._stp_params["system-id"]
    except KeyError:
      return None

  @address.setter
  def address (self, val):
    self._stp_params["system-id"] = val
    self._rstp_params["address"] = val

  @property
  def ageing_time (self):
    if self._mode != OVSL2STP.RSTP:
      raise IllegalModeForParamError("ageing-time")
    try:
      return self._rstp_params["ageing-time"]
    except KeyError:
      return None

  @ageing_time.setter
  def ageing_time (self, val):
    if self._mode != OVSL2STP.RSTP:
      raise IllegalModeForParamError("ageing-time")
    self._rstp_params["ageing-time"] = val
    
  @property
  def xmit_hold_count (self):
    if self._mode != OVSL2STP.RSTP:
      raise IllegalModeForParamError("xmit-hold-count")
    try:
      return self._rstp_params["xmit-hold-count"]
    except KeyError:
      return None

  @xmit_hold_count.setter
  def xmit_hold_count (self, val):
    if self._mode != OVSL2STP.RSTP:
      raise IllegalModeForParamError("xmit-hold-count")
    self._rstp_params["xmit-hold-count"] = val

  def _write (self, element):
    se = ET.SubElement(element, "{%s}stp" % (Namespaces.VTS))

    if self._mode == OVSL2STP.STP:
      se.attrib["type"] = "stp"
      for k,v in self._stp_params.items():
        pe = ET.SubElement(se, "{%s}%s" % (Namespaces.VTS, k))
        pe.attrib["value"] = str(v)
    elif self._mode == OVSL2STP.RSTP:
      se.attrib["type"] = "rstp"
      for k,v in self._rstp_params.items():
        pe = ET.SubElement(se, "{%s}%s" % (Namespaces.VTS, k))
        pe.attrib["value"] = str(v)
    elif self._mode == -1:
      se.attrib["type"] = "disabled"

    return element

  def _as_jsonable (self):
    obj = {"type" : self.type}
    if self._mode == OVSL2STP.STP:
      for k,v in self._stp_params.items():
        obj[k] = v
    elif self._mode == OVSL2STP.RSTP:
      for k,v in self._rstp_params.items():
        obj[k] = v
    return obj

OVSL2STP.system_id = OVSL2STP.address

class OVSL2Image(OVSImage):
  def __init__ (self):
    super(OVSL2Image, self).__init__("bss:ovs-201")
    self.stp = OVSL2STP()
    self.mac_table_size = None
    self.mac_age = None

  def _write (self, element):
    i = super(OVSL2Image, self)._write(element)
    self.stp._write(i)
    mpe = ET.SubElement(i, "{%s}mac-table-params" % (Namespaces.VTS))
    if self.mac_table_size:
      mse = ET.SubElement(mpe, "{%s}max-size" % (Namespaces.VTS))
      mse.attrib["value"] = str(self.mac_table_size)
    if self.mac_age:
      mae = ET.SubElement(mpe, "{%s}max-age" % (Namespaces.VTS))
      mae.attrib["value"] = str(self.mac_age)
      
    return i



##################
# Image Features #
##################

class SFlow(object):
  def __init__ (self, collector_ip):
    self.collector_ip = collector_ip
    self.collector_port = 6343
    self.header_bytes = 128
    self.sampling_n = 64
    self.polling_secs = 5

  def _write (self, element):
    s = ET.SubElement(element, "{%s}sflow" % (Namespaces.VTS.name))
    s.attrib["collector"] = "%s:%d" % (self.collector_ip, self.collector_port)
    s.attrib["header-bytes"] = str(self.header_bytes)
    s.attrib["sampling-n"] = str(self.sampling_n)
    s.attrib["polling-secs"] = str(self.polling_secs)
    return s


class NetFlow(object):
  def __init__ (self, collector_ip):
    self.collector_ip = collector_ip
    self.collector_port = 6343
    self.timeout = 20

  def _write (self, element):
    s = ET.SubElement(element, "{%s}netflow" % (Namespaces.VTS))
    s.attrib["collector"] = "%s:%d" % (self.collector_ip, self.collector_port)
    s.attrib["timeout"] = str(self.timeout)
    return s

class MirrorPort(object):
  def __init__ (self, port):
    self.target = port.client_id

  def _write (self, element):
    s = ET.SubElement(element, "{%s}mirror" % (Namespaces.VTS))
    s.attrib["target"] = self.target
    return s

##################
# Graph Elements #
##################

class SSLVPNFunction(Resource):
  def __init__ (self, client_id):
    super(SSLVPNFunction, self).__init__()
    self.client_id = client_id
    self.protocol = None

  def _write (self, element):
    d = ET.SubElement(element, "{%s}function" % (Namespaces.VTS.name))
    d.attrib["client_id"] = self.client_id
    d.attrib["type"] = "sslvpn"
    return d

Request.EXTENSIONS.append(("SSLVPNFunction", SSLVPNFunction))
Request.EXTENSIONS.append(("L2SSLVPNServer", SSLVPNFunction))

class L2SSLVPNClient(Resource):
  def __init__ (self, client_id):
    super(L2SSLVPNClient, self).__init__()
    self.client_id = client_id
    self.protocol = "udp"
    self.remote_ip = None
    self.remote_port = None
    self.note = None
    self.key = None

  def _write (self, element):
    d = ET.SubElement(element, "{%s}function" % (Namespaces.VTS))
    d.attrib["type"] = "sslvpn-client"
    d.attrib["client_id"] = self.client_id
    d.attrib["remote-ip"] = str(self.remote_ip)
    d.attrib["remote-port"] = str(self.remote_port)
    d.attrib["note"] = str(self.note)
    d.text = str(self.key)
    return d

Request.EXTENSIONS.append(("L2SSLVPNClient", L2SSLVPNClient))


class Datapath(Resource):
  def __init__ (self, image, client_id):
    super(Datapath, self).__init__()
    if not isinstance(image, DatapathImage):
      raise BadImageTypeError(str(DatapathImage))
    self.image = image
    self.ports = []
    self.client_id = client_id

  @property
  def name (self):
    return self.client_id

  @name.setter
  def name (self, val):
    self.client_id = val

  def attachPort (self, port):
    if port.client_id is None:
      if port.name is None:
        port.client_id = "%s:%d" % (self.name, len(self.ports))
      else:
        port.client_id = "%s:%s" % (self.name, port.name)
    self.ports.append(port)
    return port

  def connectCrossSliver (self, other_dp):
    port = InternalCircuit(None, None, None, None)
    self.attachPort(port)
    port.target = other_dp.client_id
    return port

  def _write (self, element):
    d = ET.SubElement(element, "{%s}datapath" % (Namespaces.VTS.name))
    d.attrib["client_id"] = self.name
    self.image._write(d)
    for port in self.ports:
      port._write(d)
    return d

Request.EXTENSIONS.append(("Datapath", Datapath))


class Container(Resource):
  EXTENSIONS = []

  def __init__ (self, image, name):
    super(Container, self).__init__()
    self.image = image
    self.ports =[]
    self.name = name
    self.ram = None
    self.routes = []

    for name,ext in Container.EXTENSIONS:
      self._wrapext(name, ext)

  def attachPort (self, port):
    if port.name is None:
      port.client_id = "%s:%d" % (self.name, len(self.ports))
    else:
      port.client_id = "%s:%s" % (self.name, port.name)
    self.ports.append(port)
    return port

  def addIPRoute (self, network, gateway):
    self.routes.append((ipaddress.IPv4Network(unicode(network)), ipaddress.IPv4Address(unicode(gateway))))

  def connectCrossSliver (self, other_dp):
    port = InternalCircuit(None, None, None, None)
    self.attachPort(port)
    port.target = other_dp.client_id
    return port

  def _write (self, element):
    d = ET.SubElement(element, "{%s}container" % (Namespaces.VTS.name))
    d.attrib["client_id"] = self.name
    if self.ram: d.attrib["ram"] = str(self.ram)
    self.image._write(d)
    for port in self.ports:
      port._write(d)
    for net,gw in self.routes:
      re = ET.SubElement(d, "{%s}route" % (Namespaces.VTS))
      re.attrib["network"] = str(net)
      re.attrib["gateway"] = str(gw)
    super(Container, self)._write(d)
    return d

Request.EXTENSIONS.append(("Container", Container))


class Port(object):
  def __init__ (self, name = None):
    self.client_id = None
    self.name = name

  def _write (self, element):
    p = ET.SubElement(element, "{%s}port" % (Namespaces.VTS.name))
    p.attrib["client_id"] = self.client_id
    return p


class PGCircuit(Port):
  def __init__ (self, name = None, delay_info = None):
    super(PGCircuit, self).__init__(name)
    self.delay_info = delay_info

  def _write (self, element):
    p = super(PGCircuit, self)._write(element)
    p.attrib["type"] = "pg-local"
    if self.delay_info:
      self.delay_info._write(p)
    return p

LocalCircuit = PGCircuit

class VFCircuit(Port):
  def __init__ (self, target):
    super(VFCircuit, self).__init__()
    self.target = target

  def _write (self, element):
    p = super(VFCircuit, self)._write(element)
    p.attrib["type"] = "vf-port"
    t = ET.SubElement(p, "{%s}target" % (Namespaces.VTS.name))
    t.attrib["remote-clientid"] = self.target
    return p


class InternalCircuit(Port):
  def __init__ (self, target, vlan = None, delay_info = None, loss_info = None):
    super(InternalCircuit, self).__init__()
    self.vlan = vlan
    self.target = target
    self.delay_info = delay_info
    self.loss_info = loss_info
    self.reorder_info = None

  def _write (self, element):
    p = super(InternalCircuit, self)._write(element)
    p.attrib["type"] = "internal"
    if self.vlan:
      p.attrib["vlan-id"] = str(self.vlan)
    if self.delay_info: self.delay_info._write(p)
    if self.loss_info: self.loss_info._write(p)
    if self.reorder_info: self.reorder_info._write(p)
    t = ET.SubElement(p, "{%s}target" % (Namespaces.VTS.name))
    t.attrib["remote-clientid"] = self.target
    return p


class ContainerPort(InternalCircuit):
  def __init__ (self, target, vlan = None, delay_info = None, loss_info = None):
    super(ContainerPort, self).__init__(target, vlan, delay_info, loss_info)
    self._v4addresses = []

  def _write (self, element):
    p = super(ContainerPort, self)._write(element)
    for addr in self._v4addresses:
      ae = ET.SubElement(p, "{%s}ipv4-address" % (Namespaces.VTS))
      ae.attrib["value"] = str(addr)
    return p

  def addIPv4Address (self, value):
    self._v4addresses.append(ipaddress.IPv4Interface(unicode(value)))


class GRECircuit(Port):
  def __init__ (self, circuit_plane, endpoint):
    super(GRECircuit, self).__init__()
    self.circuit_plane = circuit_plane
    self.endpoint = endpoint

  def _write (self, element):
    p = super(GRECircuit, self)._write(element)
    p.attrib["type"] = "gre"
    p.attrib["circuit-plane"] = self.circuit_plane
    p.attrib["endpoint"] = self.endpoint
    return p


######################
# Element Extensions #
######################

class Mount(Resource):
  def __init__ (self, type, name, mount_path):
    self.type = type
    self.name = name
    self.mount_path = mount_path
    self.attrs = {}

  def _write (self, element):
    melem = ET.SubElement(element, "{%s}mount" % (Namespaces.VTS))
    melem.attrib["type"] = self.type
    melem.attrib["name"] = self.name
    melem.attrib["path"] = self.mount_path
    for k,v in self.attrs.iteritems():
      melem.attrib[k] = str(v)
    return melem

Container.EXTENSIONS.append(("Mount", Mount))


class HgMount(Mount):
  """ Clone a public mercurial repo on a host

  Args:
    name (str): a reference name given on the mounting AM, must be unique within a sliver
    source (str): the URL to the source of repository
    mount_path (str): the path where the repository would be mounted in the host filesystem
    branch (str): the branch of the repository to be cloned on host (if any)
  """
  def __init__ (self, name, source, mount_path, branch = "default"):
    super(HgMount, self).__init__("hg", name, mount_path)
    self.attrs["source"] = source
    self.attrs["branch"] = branch

Container.EXTENSIONS.append(("HgMount", HgMount))


class SecureHgMount(Mount):
  def __init__ (self, getversion_output, name, source, mount_path, branch = "default"):
    super(SecureHgMount, self).__init__("hg-secure", name, mount_path)
    self._source = source
    self.attrs["source"] = _am_encrypt(getversion_output, source)
    self.attrs["branch"] = branch

  def rebind (self, getversion_output):
    self.attrs["source"] = _am_encrypt(getversion_output, self._source)

Container.EXTENSIONS.append(("SecureHgMount", SecureHgMount))


class DropboxMount(Mount):
  def __init__ (self, name, mount_path):
    super(DropboxMount, self).__init__("dropbox", name, mount_path)

Container.EXTENSIONS.append(("DropboxMount", DropboxMount))


#############
# Utilities #
#############

def connectInternalCircuit (dp1, dp2, delay_info = None, loss_info = None):
  dp1v = None
  dp2v = None

  if isinstance(dp1, tuple):
    dp1v = dp1[1]
    dp1 = dp1[0]

  if isinstance(dp2, tuple):
    dp2v = dp2[1]
    dp2 = dp2[0]

  if isinstance(dp1, Container):
    sp = ContainerPort(None, dp1v, delay_info, loss_info)
  elif isinstance(dp1, Datapath):
    sp = InternalCircuit(None, dp1v, delay_info, loss_info)

  if isinstance(dp2, Container):
    dp = ContainerPort(None, dp2v, delay_info, loss_info)
  elif isinstance(dp2, Datapath):
    dp = InternalCircuit(None, dp2v, delay_info, loss_info)

  dp1.attachPort(sp)
  dp2.attachPort(dp)

  sp.target = dp.client_id
  dp.target = sp.client_id

  return (sp, dp)



# Copyright (c) 2015-2017  Barnstormer Softworks, Ltd.

#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at http://mozilla.org/MPL/2.0/.

# Streamlined implementation of xmlrpc calls to AM API v2-compliant aggregates
# Only uses python requests module, without a ton of extra SSL dependencies

from __future__ import absolute_import

import xmlrpclib

from .util import _rpcpost

# pylint: disable=unsubscriptable-object
def getversion (url, root_bundle, cert, key, options = None):
  if not options: options = {}
  req_data = xmlrpclib.dumps((options,), methodname="GetVersion")
  return _rpcpost(url, req_data, (cert, key), root_bundle)

def listresources (url, root_bundle, cert, key, cred_strings, options = None, sliceurn = None):
  if not options: options = {}
  opts = {"geni_rspec_version" : {"version" : "3", "type" : "GENI"},
          "geni_available" : False,
          "geni_compressed" : False}

  if sliceurn:
    opts["geni_slice_urn"] = sliceurn

  # Allow all options to be overridden by the caller
  opts.update(options)

  req_data = xmlrpclib.dumps((cred_strings, opts), methodname="ListResources")
  return _rpcpost(url, req_data, (cert, key), root_bundle)

def deletesliver (url, root_bundle, cert, key, creds, slice_urn, options = None):
  if not options: options = {}
  req_data = xmlrpclib.dumps((slice_urn, creds, options), methodname="DeleteSliver")
  return _rpcpost(url, req_data, (cert, key), root_bundle)

def sliverstatus (url, root_bundle, cert, key, creds, slice_urn, options = None):
  if not options: options = {}
  req_data = xmlrpclib.dumps((slice_urn, creds, options), methodname="SliverStatus")
  return _rpcpost(url, req_data, (cert, key), root_bundle)

def renewsliver (url, root_bundle, cert, key, creds, slice_urn, date, options = None):
  FMT = "%Y-%m-%dT%H:%M:%S+00:00"
  if not options: options = {}
  req_data = xmlrpclib.dumps((slice_urn, creds, date.strftime(FMT), options), methodname="RenewSliver")
  return _rpcpost(url, req_data, (cert, key), root_bundle)

def listimages (url, root_bundle, cert, key, cred_strings, owner_urn, options = None):
  if not options: options = {}
  req_data = xmlrpclib.dumps((owner_urn, cred_strings, options), methodname="ListImages")
  return _rpcpost(url, req_data, (cert, key), root_bundle)

def createsliver (url, root_bundle, cert, key, creds, slice_urn, rspec, users, options = None):
  if not options: options = {}
  req_data = xmlrpclib.dumps((slice_urn, creds, rspec, users, options), methodname="CreateSliver")
  return _rpcpost(url, req_data, (cert, key), root_bundle)




# Copyright (c) 2014-2015  Barnstormer Softworks, Ltd.

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from __future__ import absolute_import

from lxml import etree as ET

import geni.namespaces as GNS
from .pgad import Location

TOPO = GNS.Namespace("topo", "http://geni.bssoftworks.com/rspec/ext/topo/1")

XPNS = {'g' : GNS.REQUEST.name, 'o' : GNS.OFv3.name, 't' : TOPO.name}

class Attachment(object):
  def __init__ (self):
    self.description = None

class OFAttachment(Attachment):
  def __init__ (self):
    super(OFAttachment, self).__init__()
    self.remote_cmid = None
    self.remote_port_name = None

  @classmethod
  def _fromdom (cls, elem):
    a = OFAttachment()
    a.description = elem.get("desc")
    a.remote_cmid = elem.get("remote-component-id")
    a.remote_port_name = elem.get("remote-port-name")
    return a


class Port(object):
  def __init__ (self):
    self.name = None
    self.number = None
    self.topo = []

  def __repr__ (self):
    return "<Port %s,%s>" % (self.name, self.number)

  @classmethod
  def _fromdom (cls, elem):
    p = Port()
    p.name = elem.get("name")
    p.number = elem.get("number")
    for ofa in elem.xpath('t:geni-of', namespaces = XPNS):
      p.topo.append(OFAttachment._fromdom(ofa))
    return p


class Datapath(object):
  def __init__ (self):
    self.dpid = None
    self.component_id = None
    self.ports = []
    self.location = None

  @classmethod
  def _fromdom (cls, elem):
    d = Datapath()
    d.dpid = elem.get("dpid")
    d.component_id = elem.get("component_id")

    ports = elem.xpath('o:port', namespaces = XPNS)
    for pelem in ports:
      d.ports.append(Port._fromdom(pelem))

    lelem = elem.xpath('o:location', namespaces = XPNS)
    if lelem:
      d.location = Location._fromdom(lelem[0])

    return d


class Advertisement(object):
  def __init__ (self, path = None, xml = None):
    if path:
      self._root = ET.parse(open(path))
    elif xml:
      self._root = ET.fromstring(xml)

  @property
  def text (self):
    return ET.tostring(self._root, pretty_print=True)

  @property
  def datapaths (self):
    for datapath in self._root.findall("{%s}datapath" % (GNS.OFv3.name)):
      yield Datapath._fromdom(datapath)



# Copyright (c) 2016 The University of Utah

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from __future__ import absolute_import

from lxml import etree as ET

from ..pg import Node, Link, Request, Namespaces

class UserDataSet(object):
  def __init__ (self, data=None, prefix="emulab.net.userdata."):
    if data is not None:
      self.data = data
    else:
      self.data = {}
    self.prefix = prefix

  def addData (self, element, data=None):
    if isinstance(element, dict):
      self.data.update(element)
    elif data is not None:
      self.data[element] = data


class UserData(object):
  def __init__ (self, userdata):
    self.userdata = None
    if isinstance(userdata, UserDataSet):
      self.userdata = userdata

  def _write (self, element):
    if self.userdata is not None:
      ud = ET.SubElement(element, "data_set",
                         nsmap={None : Namespaces.PARAMS.name})
      for key, value in self.userdata.data.iteritems():
        udi = ET.SubElement(ud, "data_item")
        udi.attrib["name"] = self.userdata.prefix + key

        if isinstance(value, ET._Element):
          udi.append(value)
        else:
          udi.text = value
      return ud

Node.EXTENSIONS.append(("UserData", UserData))
Link.EXTENSIONS.append(("UserData", UserData))



# Copyright (c) 2016 The University of Utah

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from __future__ import absolute_import

from ..pg import LAN, Request
from .pndefs import EPCLANS

#
# EPC lan class.  Convenience wrapper.
#
class EPClan(LAN):

    def __init__(self, name, vmlan = False):
        super(EPClan, self).__init__(name)
        if vmlan:
            self.vlan_tagging = 1
            self.trivial_ok = 1
            self.link_multiplexing = 1
        # Force mgmt lan to be best effort.
        if name == EPCLANS.MGMT:
            self.bandwidth = -1
            self.best_effort = 1

    def isMember(self, node):
        for intf in self.interfaces:
            if intf.node == node:
                return True
        return False

    def addMember(self, node, bandwidth = 0, latency = 0, plr = 0):
        intf = node.addInterface(self.client_id)
        if bandwidth:
            intf.bandwidth = bandwidth
        if latency:
            intf.latency = latency
        if plr:
            intf.plr = plr
        self.addInterface(intf)
        return intf

    def _write(self, root):
        if not self.bandwidth:
            self.bandwidth = -1
            self.best_effort = 1
        return super(EPClan, self)._write(root)

Request.EXTENSIONS.append(("EPClan", EPClan))



# Copyright (c) 2015-2016  Barnstormer Softworks, Ltd.

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""
Tempfile Manager to enable secure creation of temporary files, and deletion on process exit.

The module-level `makeFile` method uses a pre-existing singleton manager for the entire process,
while individual `TempfileManager` instances can be created to serve different user needs and scopes.
"""

from __future__ import absolute_import

import tempfile
import os
import shutil
import atexit

class TempfileManager(object):
  """Global tempfile manager for the current process that guarantees deletion when the process ends."""

  def __init__ (self):
    self.path = tempfile.mkdtemp()

  def clear (self):
    shutil.rmtree(self.path, ignore_errors = True)

  def makeFile (self):
    """Create a temporary file.  The open file object and full path are returned.
       .. note:
         You may delete this file at any time, otherwise it will be deleted when the process exits."""
    (handle, path) = tempfile.mkstemp(dir = self.path)
    return (os.fdopen(handle, "w"), path)

TFM = TempfileManager()

makeFile = TFM.makeFile

atexit.register(TFM.clear)



# Copyright (c) 2013-2017  Barnstormer Softworks, Ltd.

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from __future__ import absolute_import

from lxml import etree as ET

from .. import namespaces as GNS
from .pg import Namespaces as PGNS
from . import pg
from . import stitching
from ..model.util import XPathXRange

_XPNS = {'g' : GNS.REQUEST.name, 's' : GNS.SVLAN.name,
         'e' : PGNS.EMULAB.name, 't' : stitching.STITCHNS.name}

class Image(object):
  def __init__ (self):
    self.name = None
    self.os = None
    self.version = None
    self.description = None
    self.url = None

  def __repr__ (self):
    return "<Image: %s, os: '%s', version: '%s', description: '%s', url: '%s'>" % (self.name, self.os, self.version,
                                                                                   self.description, self.url)

  def __hash__ (self):
    return hash("%s-%s" % (self.name, self.url))

  def __eq__ (self, other):
    return hash(self) == hash(other)

  def __ne__ (self, other):
    return not self == other

  @classmethod
  def _fromdom (cls, elem):
    i = Image()
    i.name = elem.get("name")
    if i.name is None:
      i.name = elem.get("url")
    i.os = elem.get("os")
    i.version = elem.get("version")
    i.description = elem.get("description")
    i.url = elem.get("url")
    return i


class Location(object):
  def __init__ (self):
    self.latitude = None
    self.longitude = None

  def __repr__ (self):
    return "<Location: %f, %f>" % (self.latitude, self.longitude)

  @classmethod
  def _fromdom (cls, elem):
    l = Location()
    l.latitude = float(elem.get("latitude"))
    l.longitude = float(elem.get("longitude"))
    return l


class AdInterface(pg.Interface):
  """Wrapper object for a Node Interface in a GENIv3 Advertisement.

  Attributes:
    component_id (str): Component ID URN
    role (str): The resource role of this interface (typically
      "control" or "experimental").  `None` if unset.
    name (str): Friendly name for this interface, `None` if unset.
  """
  def __init__ (self, name):
    super(AdInterface, self).__init__(name, None)
    self.component_id = None
    self.role = None

  @classmethod
  def _fromdom (cls, elem):
    eie = elem.xpath('e:interface', namespaces = _XPNS)
    name = elem.get("component_id")
    if len(eie) > 0:
      name = eie[0].get("name")
    intf = AdInterface(name)

    intf.component_id = elem.get("component_id")
    intf.role = elem.get("role")

    return intf


class AdNode(object):
  """Wrapper object for a Node in a GENIv3 advertisement.

  .. note::
    In general this object is created on-demand through `Advertisement` objects,
    but you can load this object from a Node XML element by using the `_fromdom`
    classmethod.

    Attributes:
      component_id (str): Component ID URN
      component_manager_id (str): Component Manager ID URN
      name (str): Friendly name provided by aggregate for this resource.
      exclusive (bool): True if a node can be reserved as a raw PC
      available (bool): Whether this node is currently available for reservations
      hardware_types (dict): Mapping of `{ type_name : type_slots, ... }`
      sliver_types (set): Supported sliver type
      images (dict): Mapping of `{ sliver_type : [supported_image_name, ...], ... }`
      shared (bool): `True` if currently being used as a shared resource
      interfaces (list): List of :py:class:`AdInterface` objects for this Node
      location (:py:class:`AdLocation`): `None` if not available
      ram (int): Currently available system RAM in megabytes.  `None` if not available.
      cpu (int): Maximum Per-core CPU speed in Mhz.  `None` if not available.
  """

  def __init__ (self):
    self.component_id = None
    self.component_manager_id = None

    self.name = None
    self.exclusive = True
    self.available = False
    self.hardware_types = {}
    self.sliver_types = set()
    self.images = {}
    self.shared = False
    self.interfaces = []
    self.location = None
    self.ram = None
    self.cpu = None
    self._elem = None

  @classmethod
  def _fromdom (cls, elem):
    node = AdNode()
    node._elem = elem
    node.component_id = elem.get("component_id")
    node.name = elem.get("component_name")
    node.component_manager_id = elem.get("component_manager_id")
    if elem.get("exclusive") == "false":
      node.exclusive = False

    avelem = elem.xpath('g:available', namespaces = _XPNS)
    if avelem and avelem[0].get("now") == "true":
      node.available = True

    stypes = elem.xpath('g:sliver_type', namespaces = _XPNS)
    for stype in stypes:
      sliver_name = stype.get("name")
      node.sliver_types.add(sliver_name)
      node.images[sliver_name] = []
      ims = stype.xpath('g:disk_image', namespaces = _XPNS)
      for im in ims:
        node.images[sliver_name].append(Image._fromdom(im))

    htypes = elem.xpath('g:hardware_type', namespaces = _XPNS)
    for htype in htypes:
      nts = htype.xpath('e:node_type', namespaces = _XPNS)
      if nts:
        node.hardware_types[htype.get("name")] = nts[0].get("type_slots")

    fds = elem.xpath('e:fd', namespaces = _XPNS)
    for fd in fds:
      name = fd.get("name")
      if name == 'pcshared':
        node.shared = True
      elif name == 'cpu':
        node.cpu = int(fd.get("weight"))
      elif name == 'ram':
        node.ram = int(fd.get("weight"))

    for intf in elem.xpath('g:interface', namespaces = _XPNS):
      node.interfaces.append(AdInterface._fromdom(intf))

    locelem = elem.xpath('g:location', namespaces = _XPNS)
    if locelem:
      node.location = Location._fromdom(locelem[0])

    return node

  @property
  def text (self):
    return ET.tostring(self._elem, pretty_print=True)


class AdLink(object):
  def __init__ (self):
    self.component_id = None
    self.link_types = set()
    self._elem = None
    self.interface_refs = []

  @classmethod
  def _fromdom (cls, elem):
    link = AdLink()
    link._elem = elem
    link.component_id = elem.get("component_id")

    ltypes = elem.xpath('g:link_type', namespaces = _XPNS)
    for ltype in ltypes:
      link.link_types.add(ltype.get("name"))

    irefs = elem.xpath('g:interface_ref', namespaces = _XPNS)
    for iref in irefs:
      link.interface_refs.append(iref.get("component_id"))

    return link

  @property
  def text (self):
    return ET.tostring(self._elem, pretty_print=True)


class AdSharedVLAN(object):
  def __init__ (self):
    self.name = None

  def __str__ (self):
    return self.name

  @classmethod
  def _fromdom (cls, elem):
    svlan = AdSharedVLAN()
    svlan.name = elem.get("name")
    return svlan


class RoutableAddresses(object):
  def __init__ (self):
    self.available = 0
    self.configured = 0

  @property
  def capacity (self):
    return self.configured


class Advertisement(object):
  """Wrapper object for a GENIv3 XML advertisement.

  Only one argument can be supplied (if both are provided `path` will be used)

  Args:
    path (str, unicode): Path to XML file on disk containing an advertisement
    xml (str, unicode): In-memory XML byte stream containing an advertisement
  """

  def __init__ (self, path = None, xml = None):
    if path:
      self._root = ET.parse(open(path))
    elif xml:
      self._root = ET.fromstring(xml)
    self._routable_addresses = None
    self._images = set()

  def _parse_routable (self):
    try:
      elem = self._root.xpath('/g:rspec/e:rspec_routable_addresses', namespaces=_XPNS)[0]
      ra = RoutableAddresses()
      ra.available = int(elem.get("available"))
      ra.configured = int(elem.get("configured"))
      self._routable_addresses = ra
    except Exception:
      pass

  @property
  def routable_addresses (self):
    """A RoutableAddresses object containing the number of configured and available publicly routable IP addresses at this site."""
    if not self._routable_addresses:
      self._parse_routable()
    return self._routable_addresses

  @property
  def nodes (self):
    """An indexable iterator over the AdNode objects in this advertisement."""
    return XPathXRange(self._root.findall("{%s}node" % (GNS.REQUEST.name)), AdNode)

  @property
  def links (self):
    """An indexable iterator over the AdLink objects in this advertisement."""
    return XPathXRange(self._root.findall("{%s}link" % (GNS.REQUEST.name)), AdLink)

  @property
  def shared_vlans (self):
    """An indexable iterator of the shared vlan names found in this advertisement."""
    return XPathXRange(self._root.xpath('/g:rspec/s:rspec_shared_vlan/s:available', namespaces=_XPNS), AdSharedVLAN)

  @property
  def images (self):
    """An iterable of the unique images found in this advertisement."""
    if not self._images:
      for node in self.nodes:
        for image_list in node.images.values():
          for image in image_list:
            self._images.add(image)
    for image in self._images:
      yield image

  @property
  def stitchinfo (self):
    """Reference to the stitching info in the manifest, if present."""
    try:
      elem = self._root.xpath('/g:rspec/t:stitching', namespaces=_XPNS)[0]
      return stitching.AdInfo(elem)
    except IndexError:
      return None

  @property
  def text (self):
    """Advertisement XML contents as a string, formatted with whitespace for easier reading."""
    return ET.tostring(self._root, pretty_print=True)

  def writeXML (self, path):
    """Write the current advertisement as an XML file that contains an rspec in the format returned by the
    aggregate."""
    f = open(path, "w+")
    f.write(ET.tostring(self._root, pretty_print=True))
    f.close()



# Copyright (c) 2014-2015  Barnstormer Softworks, Ltd.

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.


from __future__ import absolute_import

from . import RSpec

class Request(RSpec):
  def __init__ (self):
    super(Request, self).__init__("request")
    self._resources = []

  def addGroup (self, name):
    group = Group(name)
    self._resources.append(group)
    return group

  def buildMatch (self):
    match = Match()
    self._resources.append(match)
    return match

  @property
  def resources(self):
    return self._resources

class Datapath(object):
  def __init__ (self, ad_dp = None):
    self._ad_dp = ad_dp



# Copyright (c) 2014-2017 Barnstomer Softworks, Ltd. and The University of Utah

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from __future__ import absolute_import

import inspect
import sys

from .protogeni import PGCompute

# Imports to pick up aggregates available in the cloudlab UI
from .instageni import UtahDDC # pylint: disable=unused-import
from .apt import Apt # pylint: disable=unused-import

class CloudLabAM(PGCompute): pass


Clemson = CloudLabAM("cl-clemson", "boss.clemson.cloudlab.us", "urn:publicid:IDN+clemson.cloudlab.us+authority+cm")
Utah = CloudLabAM("cl-utah", "boss.utah.cloudlab.us", "urn:publicid:IDN+utah.cloudlab.us+authority+cm")
Wisconsin = CloudLabAM("cl-wisconsin", "www.wisc.cloudlab.us", "urn:publicid:IDN+wisc.cloudlab.us+authority+cm")

def aggregates ():
  module = sys.modules[__name__]
  for _,obj in inspect.getmembers(module):
    if isinstance(obj, PGCompute):
      yield obj

def name_to_aggregate ():
  result = dict()
  module = sys.modules[__name__]
  for _,obj in inspect.getmembers(module):
    if isinstance(obj, PGCompute):
      result[obj.name] = obj
  return result



# Copyright (c) 2014-2017 The University of Utah and Barnstormer Softworks, Ltd.

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""Simple library for manipulating URNs, particularly those used for GENI
objects"""

from __future__ import absolute_import

import re

from geni.aggregate.core import AM
from geni.exceptions import WrongNumberOfArgumentsError

def Make(s):
  """Returns the 'most specific' URN object that it can for the given string.

  Specifically, returns a GENI URN if the string is in GENI format, or a Base
  URN if it is not.  May throw a MalformedURNError exception if the string is not a
  valid URN at all."""
  if GENI.isValidGENIURN(s):
    return GENI(s)
  return Base(s)

class MalformedURNError(Exception):
  """Exception indicating that a string is not a proper URN."""

  def __init__ (self, val):
    super(MalformedURNError, self).__init__()
    self._val = val

  def __str__(self):
    return "Malformed URN: %s" % self._val

class Base (object):
  """Base class representing any URN (RFC 2141)."""

  PREFIX = "urn"

  NID_PATTERN = "[a-z0-9][a-z0-9-]{0,31}"
  NSS_PATTERN = r"""[a-z0-9()+,\-.:=@;$_!*'%/?#]+"""
  NID_REGEX = re.compile("^%s$" % NID_PATTERN, re.IGNORECASE)
  NSS_REGEX = re.compile("^%s$" % NSS_PATTERN, re.IGNORECASE)
  URN_REGEX = re.compile("^urn:%s:%s$" % (NID_PATTERN, NSS_PATTERN),
                         re.IGNORECASE)

  @staticmethod
  def isValidURN(s):
    """Returns True if the string is a valid URN, False if not."""
    return Base.URN_REGEX.match(s) is not None

  @staticmethod
  def isValidNID(s):
    """Returns True if the string is a valid NID, False if not."""
    return Base.NID_REGEX.match(s) is not None

  @staticmethod
  def isValidNSS(s):
    """Returns True if the string is a valid NSS, False if not."""
    return Base.NSS_REGEX.match(s) is not None

  @staticmethod
  def _fromStr(s):
    if not Base.isValidURN(s):
      raise MalformedURNError(s)
    return tuple(re.split(":",s,2))

  def __init__ (self, *args):
    """Create a new generic URN

    URNs can be initialized in one of two ways:

    1) Passing a single string in URN format ('urn:NID:NSS')

    2) Passing two strings (the NID and the NSS) separately"""
    if len(args) == 1:
      # Note, _fromStr will thrown an exception if malformed
      (_, self._nid, self._nss) = Base._fromStr(args[0])
    elif len(args) == 2:
      if not Base.isValidNID(args[0]):
        raise MalformedURNError("NID: %s" % args[0])
      if not Base.isValidNSS(args[1]):
        raise MalformedURNError("NSS: %s" % args[1])
      self._nid = args[0]
      self._nss = args[1]
    else:
      raise WrongNumberOfArgumentsError()

  def __str__ (self):
    return "%s:%s:%s" % (Base.PREFIX, self._nid, self._nss)

  __repr__ = __str__

class GENI (Base):
  """Class representing the URNs used by GENI, which use the publicid NID and
  IDN (domain name) scheme, then impose some additional strucutre."""

  NID = "publicid"
  NSSPREFIX = "IDN"

  # Taken from the list at http://www.protogeni.net/wiki/URNs
  TYPE_AUTHORITY = "authority"  #: Aggregate Managers, Slice Authorities, etc.
  TYPE_IMAGE     = "image"      #: Disk images
  TYPE_INTERFACE = "interface"  #: Network interfaces
  TYPE_LINK      = "link"       #: Point-to-point and multipoint links
  TYPE_NODE      = "node"       #: Physical and virtual machines
  TYPE_SLICE     = "slice"      #: Container for allocated resources
  TYPE_SLIVER    = "sliver"     #: Slice of a specific resource
  TYPE_USER      = "user"       #: Principal

  # We use IDN for authorities, and many identifiers turn into parts of domain
  # names, so we sort of match against DNS strings (though are somewhat more
  # permissive)
  DNS_PART          = "[a-z0-9]+[a-z0-9-]*"
  DNS_FULL          = r"""(%s.?)+""" % DNS_PART

  AUTHORITY_PATTERN = "%s(:%s)*" % (DNS_FULL, DNS_FULL)
  TYPE_PATTERN      = DNS_PART
  NAME_PATTERN      = Base.NSS_PATTERN
  GENINSS_PATTERN   = r"""%s\+%s\+(?P<type>%s)\+%s""" % (NSSPREFIX, AUTHORITY_PATTERN,
                                                         TYPE_PATTERN, NAME_PATTERN)
  GENIURN_PATTERN   = "%s:%s:%s" % (Base.PREFIX, NID, GENINSS_PATTERN)

  AUTHORITY_REGEX   = re.compile("^%s$" % AUTHORITY_PATTERN, re.IGNORECASE)
  TYPE_REGEX        = re.compile("^%s$" % TYPE_PATTERN, re.IGNORECASE)
  NAME_REGEX        = re.compile("^%s$" % NAME_PATTERN, re.IGNORECASE)
  GENINSS_REGEX     = re.compile("^%s$" % GENINSS_PATTERN, re.IGNORECASE)
  GENIURN_REGEX     = re.compile("^%s$" % GENIURN_PATTERN, re.IGNORECASE)

  def __init__ (self, *args):
    """Create a URN in the format used for GENI objects

    There are four forms of this constructor:

    1) Pass a single string in GENI URN format ('urn:publicid:IDN+auth+type+name')

    2) Pass three arguments: the authority (a single string), the type (see the
       TYPE_ variables in this class), and the object name

    3) Pass three arguments: as #2, but the authorit(ies) are passed as a list,
       with the top-level authority coming first, followed by any subauthorities

    4) Pass three arguments: as #2, but the authority is a
       geni.aggregate.core.AM object, and the authority is taken from that
       object"""
    if len(args) == 1:
      # Superclass constructor parses and sets self._nss, which we then split
      # into its constituent GENI parts. _splitNSS might throw an exception.
      super(GENI,self).__init__(args[0])
      self._authorities, self._type, self._name = GENI._splitNSS(self._nss)
    elif len(args) == 3:
      if isinstance(args[0],str):
        # They gave us a string, figure out if it might have subauthorities
        # in it
        self._authorities = GENI._splitAuthorities(args[0])
      elif isinstance(args[0], AM):
        # If given an AM, extract its authority information; accept either a
        # proper URN object or a simple string
        if isinstance(args[0].component_manager_id, GENI):
          self._authorities = args[0].component_manager_id.authorities
        else:
          self._authorities = GENI(args[0].component_manager_id).authorities
      else:
        self._authorities = args[0]

      self._type = args[1]
      self._name = args[2]

      # Check if everything we got was well formed
      for authority in self._authorities:
        if not GENI.isValidAuthority(authority):
          raise MalformedURNError("Authority: %s" % authority)
      if not GENI.isValidType(self._type):
        raise MalformedURNError("Type: %s" % self._type)
      if not GENI.isValidName(self._name):
        raise MalformedURNError("Name: %s" % self._name)

      # In this form we have to reconstruct the NSS from all of the info we just
      # collected
      super(GENI,self).__init__(GENI.NID,self._makeNSS())
    else:
      raise WrongNumberOfArgumentsError()

  @property
  def authorities(self):
    """Returns a list containing at least one authority string (the top level
    authority) and possibly additional subauthorities."""
    return self._authorities

  @property
  def authority(self):
    """Return a single string capturing the entire authority/subauthority chain"""
    return ":".join(self._authorities)

  @property
  def name(self):
    """Returns the 'name' part of a GENI URN."""
    return self._name

  @property
  def type(self):
    """Returns the 'type' part of a GENI URN."""
    return self._type

  def _makeNSS(self):
    return "%s+%s+%s+%s" % (GENI.NSSPREFIX, self.authority, self._type,
                            self._name)

  @staticmethod
  def _splitNSS(s):
    if not GENI.isValidGENINSS(s):
      raise MalformedURNError("GENI NSS: %s" % s)
    matches = re.split(r"""\+""",s,4)
    return (GENI._splitAuthorities(matches[1]), matches[2], matches[3])

  @staticmethod
  def _splitAuthorities(s):
    parts = re.split(":",s)
    for part in parts:
      if not GENI.isValidAuthority(part):
        raise MalformedURNError("GENI Authority: %s" % part)
    return re.split(":",s)

  @staticmethod
  def isValidGENINSS(s):
    return GENI.GENINSS_REGEX.match(s) is not None

  @staticmethod
  def isValidAuthority(s):
    return GENI.AUTHORITY_REGEX.match(s) is not None

  @staticmethod
  def isValidType(s):
    # Note that we don't actually check against the set of known types found
    # above, as it is not a closed set
    return GENI.TYPE_REGEX.match(s) is not None

  @staticmethod
  def isValidName(s):
    return GENI.NAME_REGEX.match(s) is not None

  @staticmethod
  def GENIURNType(s):
    """Returns the type of the object if the URN is a valid GENI URN, or
    None otherwise."""
    matches = GENI.GENIURN_REGEX.match(s)
    if matches is None:
      return None
    return matches.group("type")

  @staticmethod
  def isValidGENIURN(s):
    """Returns True if the given string is a valid URN in GENI format, False
    otherwise."""
    return GENI.GENIURNType(s) is not None

def Authority (authorities, name):
  """Create a new GENI URN with type 'authority'."""
  return GENI(authorities, GENI.TYPE_AUTHORITY, name)

def Interface (authorities, name):
  """Create a new GENI URN with type 'interface'."""
  return GENI(authorities, GENI.TYPE_INTERFACE, name)

def Image (authorities, name, version = None):
  """Create a new GENI URN with type 'image'."""
  if version is not None:
    constructed_name = "%s:%s" % (name, str(version))
  else:
    constructed_name = name
  return GENI(authorities, GENI.TYPE_IMAGE, constructed_name)

def Link (authorities, name):
  """Create a new GENI URN with type 'link'."""
  return GENI(authorities, GENI.TYPE_LINK, name)

def Node (authorities, name):
  """Create a new GENI URN with type 'node'."""
  return GENI(authorities, GENI.TYPE_NODE, name)

def Slice (authorities, name):
  """Create a new GENI URN with type 'slice'."""
  return GENI(authorities, GENI.TYPE_SLICE, name)

def Sliver (authorities, name):
  """Create a new GENI URN with type 'sliver'."""
  return GENI(authorities, GENI.TYPE_SLIVER, name)

def User (authorities, name):
  """Create a new GENI URN with type 'user'."""
  return GENI(authorities, GENI.TYPE_USER, name)

if __name__ == "__main__":
  # pylint: disable=wrong-import-position,wrong-import-order,global-statement
  # Lame unit tests
  import sys
  import geni.aggregate.instageni as IG

  errors = 0

  def check_urn (urn, value):
    global errors
    if str(urn) == value:
      sys.stdout.write("PASS")
    else:
      sys.stdout.write("FAIL")
      errors = errors + 1

    sys.stdout.write(" %s / %s\n" % (urn,value))

  def check_type(s, t):
    global errors
    urn = Make(s)
    if isinstance(urn, t):
      sys.stdout.write("PASS")
    else:
      sys.stdout.write("FAIL")
      errors = errors + 1

    sys.stdout.write(" %s / %s\n" % (urn,t.__name__))

  check_urn(Base("isbn","0553575384"),"urn:isbn:0553575384")
  check_urn(GENI("emulab.net","user","ricci"),
            "urn:publicid:IDN+emulab.net+user+ricci")
  check_urn(Base("urn:isbn:0140186255"),"urn:isbn:0140186255")
  check_urn(GENI("urn:publicid:IDN+emulab.net+user+jay"),
            "urn:publicid:IDN+emulab.net+user+jay")
  check_urn(GENI(IG.Kentucky,"user","hussam"),
            "urn:publicid:IDN+lan.sdn.uky.edu+user+hussam")
  check_urn(Image(IG.UtahDDC,"UBUNTU64-STD"),
            "urn:publicid:IDN+utahddc.geniracks.net+image+UBUNTU64-STD")
  check_urn(Image(IG.UtahDDC,"UBUNTU64-STD",42),
            "urn:publicid:IDN+utahddc.geniracks.net+image+UBUNTU64-STD:42")
  check_urn(User(IG.Clemson,"kwang"),
            "urn:publicid:IDN+instageni.clemson.edu+user+kwang")
  check_urn(GENI("wisc.cloudlab.us", "image", "ramcloud-PG0:hq6_ubuntu16.04_singlePC:0"),
            "urn:publicid:IDN+wisc.cloudlab.us+image+ramcloud-PG0:hq6_ubuntu16.04_singlePC:0")

  check_type("urn:isbn:0345371984",Base)
  check_type("urn:publicid:IDN+utahddc.geniracks.net+image+UBUNTU64-STD",GENI)
  check_type("urn:publicid:IDN+utahddc.geniracks.net+image+UBUNTU64-STD:42",GENI)

  sys.exit(errors)



# Copyright (c) 2015-2017  Barnstormer Softworks, Ltd.

#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at http://mozilla.org/MPL/2.0/.

import xmlrpclib

class HTTP(object):
  """Global configuration options for MiniGCF HTTP(S) calls."""

  TIMEOUT = 60
  """Initial response timeout.  Note that this is not the time limit on the entire
  download, just the initial server response."""

  ALLOW_REDIRECTS = False
  """Allow MiniGCF to follow HTTP redirects (301)."""

  LOG_RAW_RESPONSES = False
  """If set to a valid `(log_handle, log_level)` tuple, will write all raw responses
  (before any parsing) from AM API and CH API calls to that log at the given level."""

  LOG_RAW_REQUESTS = False
  """If set to a valid `(log_handle, log_level)` tuple, will write all raw requests
  (before any parsing) to AM API and CH API calls to that log at the given level."""

  LOG_URLS = False
  """If set to a valid `(log_handle, log_level)` tuple, will log all URLs as they are used."""



# Copyright (c) 2014  Barnstormer Softworks, Ltd.

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import json
import requests

from . import germ

class Admin(object):
  URL = "https://%s:%d/core/admin/%s"

  def __init__ (self, connection):
    self._conn = connection

  @property
  def configkeys (self):
    url = Admin.URL % (self._conn.host, self._conn.port, "get-config-keys")
    r = requests.get(url, **self._conn.rkwargs)
    return r.json()["value"]

  @property
  def version (self):
    url = Admin.URL % (self._conn.host, self._conn.port, "get-version")
    r = requests.get(url, **self._conn.rkwargs)
    return r.json()["value"]["version"]

  def getConfig (self, key):
    url = Admin.URL % (self._conn.host, self._conn.port, "get-config")
    d = json.dumps({"key":key})
    r = requests.post(url, d, **self._conn.rkwargs)
    return r.json()["value"]

class GENI(object):
  URL = "https://%s:%d/core/admin/%s"

  def __init__ (self, connection):
    self._c = connection

  @property
  def slivers (self):
    url = GENI.URL % (self._c.host, self._c.port, "list-slivers")
    r = requests.get(url, **self._c.rkwargs)
    d = r.json()
    if d["retcode"] != 0:
      return None #TODO Raise exception
    sl = []
    # TODO: Make a real sliver object that can lazily get back to type-specific data
    for sliver in d["value"]["slivers"]:
      sl.append(sliver)
    return sl


class Connection(germ.Connection):
  def __init__ (self):
    super(Connection, self).__init__()
    self.user = "foamadmin"
    self.admin = Admin(self)
    self.geni = GENI(self)



# Copyright (c) 2016-2017 The University of Utah

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""
Common set of RSpec extensions supported by many Emulab-based aggregates
"""

from __future__ import absolute_import

from ..pg import Request, Namespaces, Link, Node, Service, Command, RawPC
import geni.namespaces as GNS
from lxml import etree as ET

class setCollocateFactor(object):
    """Added to a top-level Request object, this extension limits the number
    of VMs from one experiment that Emulab will collocate on each physical
    host.
    """
    __ONCEONLY__ = True
    
    def __init__(self, mfactor):
        """mfactor is an integer, giving the maximum number of VMs to multiplex
        on each physical host."""
        self.mfactor = mfactor
    
    def _write(self, root):
        el = ET.SubElement(root,
                           "{%s}collocate_factor" % (Namespaces.EMULAB.name))
        el.attrib["count"] = str(self.mfactor)
        return root

Request.EXTENSIONS.append(("setCollocateFactor", setCollocateFactor))

class setPackingStrategy(object):
    """Added to a top-level Request object, this extension controls the
    strategy used for distributing VMs across physical hosts
    """
    __ONCEONLY__ = True

    def __init__(self, strategy):
        self.strategy = strategy
    
    def _write(self, root):
        el = ET.SubElement(root,
                           "{%s}packing_strategy" % (Namespaces.EMULAB.name))
        el.attrib["strategy"] = str(self.strategy)
        return root

Request.EXTENSIONS.append(("setPackingStrategy", setPackingStrategy))
    
class setRoutingStyle(object):
    """Added to a top-level Request object, this extension controls the
    routing that is automatically configured on the experiment (data-plane)
    side of the network.
    """
    __ONCEONLY__ = True
    
    def __init__(self, style):
        self.style = style
    
    def _write(self, root):
        el = ET.SubElement(root,
                           "{%s}routing_style" % (Namespaces.EMULAB.name))
        el.attrib["style"] = str(self.style)
        return root

Request.EXTENSIONS.append(("setRoutingStyle", setRoutingStyle))

class setDelayImage(object):
    """Added to a top-level Request object, this extension sets the disk image
    that will be used for all delay nodes configured for the experiment.
    """
    __ONCEONLY__ = True
    
    def __init__(self, urn):
        """urn: URN of any image - to perform the intnded function, the 
        image must be capable of setting up bridging and/or traffic shaping.
        """
        self.urn = urn
    
    def _write(self, root):
        el = ET.SubElement(root,
                           "{%s}delay_image" % (Namespaces.EMULAB.name))
        el.attrib["urn"] = str(self.urn)
        return root

Request.EXTENSIONS.append(("setDelayImage", setDelayImage))

class setForceShaping(object):
    """Added to a Link or LAN object, this extension forces Emulab link
    shaping to be enabled, even if it is not strictly necessary. This
    allows the link properties to be changed dynamically via the Emulab event
    system.
    """
    __ONCEONLY__ = True
    
    def __init__(self):
        self._enabled = True
    
    def _write(self, root):
        if self._enabled == False:
            return root
        el = ET.SubElement(root, "{%s}force_shaping" % (Namespaces.EMULAB.name))
        el.attrib["enabled"] = "true"
        return root

Link.EXTENSIONS.append(("setForceShaping", setForceShaping))

class setNoBandwidthShaping(object):
    """Added to a Link or LAN object, this extension forces Emulab link
    shaping to be disabled for bandwidth, even if it is necessary. This
    is ignored if the link must be shaped for other reason (delay, loss).
    """
    __ONCEONLY__ = True
    
    def __init__(self):
        self._enabled = True
    
    def _write(self, root):
        if self._enabled == False:
            return root
        el = ET.SubElement(root,
                           "{%s}force_nobwshaping" % (Namespaces.EMULAB.name))
        el.attrib["enabled"] = "true"
        return root

Link.EXTENSIONS.append(("setNoBandwidthShaping", setNoBandwidthShaping))

class setNoInterSwitchLinks(object):
    """Added to a Link or LAN object, this extension forces the Emulab mapper
    to disallow mapping a link in the request topology to an inter-switch
    link.  This allows users to require that specific nodes in their
    topology be attached to the same switch(es).
    """
    __ONCEONLY__ = True
    
    def __init__(self):
        self._enabled = True
    
    def _write(self, root):
        if self._enabled == False:
            return root
        el = ET.SubElement(root, "{%s}interswitch" % (Namespaces.EMULAB.name))
        el.attrib["allow"] = "false"
        return root

Link.EXTENSIONS.append(("setNoInterSwitchLinks", setNoInterSwitchLinks))

class setUseTypeDefaultImage(object):
    """Added to a node that does not specify a disk image, this extension
    forces Emulab to use the hardware type default image instead of the
    standard geni default image. Useful with special hardware that should
    run a special image.
    """
    __ONCEONLY__ = True
    
    def __init__(self):
        self._enabled = True
    
    def _write(self, root):
        if self._enabled == False:
            return root
        el = ET.SubElement(root, "{%s}use_type_default_image"
                           % (Namespaces.EMULAB.name))
        el.attrib["enabled"] = "true"
        return root

Node.EXTENSIONS.append(("setUseTypeDefaultImage", setUseTypeDefaultImage))

class setFailureAction(object):
    """Added to a node this extension will tell Emulab based aggregates to
    ignore errors booting this node when starting an experiment. This allows
    the experiment to proceed so that the user has time to debug."""
    __ONCEONLY__ = True
    
    def __init__(self, action):
        self.action = action
        self._enabled = True
    
    def _write(self, root):
        if self._enabled == False:
            return root
        el = ET.SubElement(root, "{%s}failure_action" % (Namespaces.EMULAB.name))
        el.attrib["action"] = self.action
        return root

Node.EXTENSIONS.append(("setFailureAction", setFailureAction))

#
# Emulab Program Agents.
#
class ProgramAgent(Service):
    """Add an Emulab Program Agent, which can be controlled via the Emulab
    event system. Optional argument 'directory' specifies where to invoke
    the command from. Optional argument 'onexpstart' says to invoke the
    command when the experiment starts (time=0 in event speak). This is
    different than the Execute service, which runs every time the node boots.
    """
    def __init__ (self, name, command, directory = None, onexpstart = False):
        super(ProgramAgent, self).__init__()
        self.name = name
        self.command = command
        self.directory = directory
        self.onexpstart = onexpstart

    def _write (self, element):
        exc = ET.SubElement(element,
                            "{%s}program-agent" % (Namespaces.EMULAB.name))
        exc.attrib["name"] = self.name
        if isinstance(self.command, Command):
            exc.attrib["command"] = self.command.resolve()
        else:
            exc.attrib["command"] = self.command
        if self.directory:
            exc.attrib["directory"] = self.directory
        if self.onexpstart:
            exc.attrib["onexpstart"] = "true"
        return exc

class InstantiateOn(object):
    """Added to a node to specify that it a Xen VM should be bound to
    (instantiated on) another node in the topology.  Argument is the
    node instance or the client id of another node in the topology.
    """
    class InvalidParent(Exception):
        def __init__ (self, parent):
            super(InstantiateOn.InvalidParent, self).__init__()
            self.parent = parent
            def __str__ (self):
                return "%s is not a Raw PC" % (self.parent.name)
    
    __ONCEONLY__ = True
    
    def __init__(self, parent):
        if isinstance(parent, Node):
            # Xen VMs have to be bound to a raw PC.
            if not isinstance(parent, RawPC):
                raise InvalidParent(parent)
            self._parent = parent.name
        else:
            # Allow plain name to be used. At the moment the NS converter
            # is not trying to order nodes, so the vhost might not be
            # first. 
            self._parent = parent
            
    def _write(self, root):
        if self._parent == None:
            return root
        el = ET.SubElement(root, "{%s}relation" % (GNS.REQUEST.name))
        el.attrib["type"] = "instantiate_on"
        el.attrib["client_id"] = self._parent
        return root

Node.EXTENSIONS.append(("InstantiateOn", InstantiateOn))

#
# A Bridged Link is syntatic sugar for two links separated by a bridge
# node acting as a delay node.
#
# Unfortunately, there is no way to get a handle on the parent object
# of an extension, so we need to get that explicitly.
#
class BridgedLink(object):
  """A bridged link is syntactic sugar used to create two links
separated by an Emulab delay (bridge) node. The BridgedLink class will
create the following topology:

	      left-link          right-link
	node1 =========== bridge ============ node2

The bridge is a special node type (sliver_type="delay") that tells the
CM to insert an Emulab delay node instead of a plain (router) node. A
delay node is a transparent Ethernet bridge between the left and right
segments above, but on which the traffic can be shaped wrt. bandwidth,
latency, and loss. For example:

        # Create the bridged link between the two nodes.
        link = request.BridgedLink("link")
        # Add two interfaces
        link.addInterface(iface1)
        link.addInterface(iface2)
        # Give the link (bridge) some shaping parameters.
        link.bandwidth = 10000
        link.latency   = 15
        link.plr       = 0.01"""

  # This tells the Request class to set the _parent member after creating the
  # object
  __WANTPARENT__ = True;
  
  def __init__ (self, name = None):
    if name is None:
      self.name = Link.newLinkID()
    else:
      self.name = name

    self.bridge_name = name + "_bridge"
    self.left_name   = name + "_left"
    self.right_name  = name + "_right"
    self.left_iface  = None
    self.right_iface = None

    self._bandwidth  = Link.DEFAULT_BW
    self._latency    = Link.DEFAULT_LAT
    self._plr        = Link.DEFAULT_PLR

    # This needs to get set with the setter; this helps us remember that we
    # have not been attached to a parent
    self.request = None

    # These will be set later when we know the parent
    self.bridge = None
    self.left_link = None
    self.right_link = None

  @property
  def _parent(self):
    return self.request

  @_parent.setter
  def _parent(self, request):
    self.request     = request
    self.bridge      = request.Bridge(self.bridge_name)
    self.left_link   = request.Link(self.left_name)
    self.left_link.addInterface(self.bridge.iface0);
    self.right_link  = request.Link(self.right_name)
    self.right_link.addInterface(self.bridge.iface1);

  def addInterface(self, interface):
      if self.left_iface == None:
          self.left_link.addInterface(interface)
          self.left_iface = interface
      else:
          self.right_link.addInterface(interface)
          self.right_iface = interface

  @property
  def bandwidth (self):
    return self._bandwidth

  @bandwidth.setter
  def bandwidth (self, val):
    self.bridge.pipe0.bandwidth = val;
    self.bridge.pipe1.bandwidth = val;
    self._bandwidth = val

  @property
  def latency (self):
    return self._latency

  @latency.setter
  def latency (self, val):
    self.bridge.pipe0.latency = val;
    self.bridge.pipe1.latency = val;
    self._latency = val

  @property
  def plr (self):
    return self._plr

  @plr.setter
  def plr (self, val):
    self.bridge.pipe0.lossrate = val;
    self.bridge.pipe1.lossrate = val;
    self._plr = val

  def _write(self, root):
      return root

Request.EXTENSIONS.append(("BridgedLink", BridgedLink))

class ShapedLink(BridgedLink):
  """A ShapedLink is a synonym for BridgedLink"""

  def __init__ (self, name = None):
    super(ShapedLink, self).__init__(name=name)

Request.EXTENSIONS.append(("ShapedLink", ShapedLink))



# Copyright (c) 2013  Barnstormer Softworks, Ltd.

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from __future__ import absolute_import

from geni.model import base

class ComputeDiskImage(base.Image):
  def __init__ (self):
    super(ComputeDiskImage, self).__init__ ()
    self.url = None
    self.os = None
    self.osver = None
    self.arch = None




# Copyright (c) 2014  Barnstormer Softworks, Ltd.

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.


def shared_xen (node):
  if not node.exclusive:
    if "emulab-xen" in node.sliver_types:
      return True
  return False



# Copyright (c) 2017  Barnstormer Softworks, Ltd.

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

# XML-RPC bindings for Protogeni CH/SA API revision 1

from __future__ import absolute_import

import xmlrpclib

from .util import _rpcpost

def ListComponents(url, root_bundle, cert, key, cred):
  req_data = xmlrpclib.dumps(({"credential" : cred},), methodname="ListComponents")
  return _rpcpost(url, req_data, (cert,key), root_bundle)

def GetCredential(url, root_bundle, cert, key, urn = None, uuid = None):
  if urn:
    req_data = xmlrpclib.dumps(({"urn" : urn, "type" : "Slice"},), methodname="GetCredential")
  elif uuid:
    req_data = xmlrpclib.dumps(({"uuid" : uuid, "type" : "Slice"},), methodname="GetCredential")
  else:
    req_data = xmlrpclib.dumps(tuple(), methodname="GetCredential")
  return _rpcpost(url, req_data, (cert,key), root_bundle)

def GetVersion(url, root_bundle, cert, key):
  req_data = xmlrpclib.dumps(tuple(), methodname="GetVersion")
  return _rpcpost(url, req_data, (cert,key), root_bundle)

def Resolve(url, root_bundle, cert, key, cred, urn, typ):
  obj = {"credential" : cred, "urn" : urn, "type" : typ}
  req_data = xmlrpclib.dumps((obj,), methodname="Resolve")
  return _rpcpost(url, req_data, (cert,key), root_bundle)

def Register(url, root_bundle, cert, key, user_cred, hrn):
  obj = {"credential" : user_cred, "hrn" : hrn, "type" : "Slice"}
  req_data = xmlrpclib.dumps((obj,), methodname="Register")
  return _rpcpost(url, req_data, (cert,key), root_bundle)



# Copyright (c) 2017  Barnstormer Softworks, Ltd.

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from __future__ import absolute_import

class SLICE_ROLE(object):
  LEAD = "LEAD"
  ADMIN = "ADMIN"
  MEMBER = "MEMBER"
  OPERATOR = "OPERATOR"
  AUDITOR = "AUDITOR"

class PROJECT_ROLE(object):
  LEAD = "LEAD"
  ADMIN = "ADMIN"
  MEMBER = "MEMBER"

class REQCTX(object):
  PROJECT = 1
  SLICE = 2
  RESOURCE = 3
  SERVICE = 4
  MEMBER = 5

class REQSTATUS(object):
  PENDING = 0
  APPROVED = 1
  CANCELLED = 2
  REJECTED = 3



# Copyright (c) 2014-2018  Barnstormer Softworks, Ltd.

#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at http://mozilla.org/MPL/2.0/.

from __future__ import absolute_import

from .core import APIRegistry
from .exceptions import AMError
from . import pgutil as ProtoGENI

# pylint: disable=multiple-statements
class AllocateError(AMError): pass
class CreateSliverError(AMError): pass
class DeleteSliverError(AMError): pass
class GetVersionError(AMError): pass
class ListResourcesError(AMError): pass
class ProvisionError(AMError): pass
class RenewSliverError(AMError): pass
class SliverStatusError(AMError): pass
class POAError(AMError): pass
# pylint: enable=multiple-statements


class AMAPIv3(object):
  @staticmethod
  def poa (context, url, sname, action, urns = None, options = None):
    from ..minigcf import amapi3 as AM3

    sinfo = context.getSliceInfo(sname)
    if not urns:
      urns = [sinfo.urn]

    res = AM3.poa(url, False, context.cf.cert, context.cf.key, [sinfo], urns, action, options)
    if res["code"]["geni_code"] == 0:
      return res["value"]

    if res["code"].has_key("am_type"):
      if res["code"]["am_type"] == "protogeni":
        ProtoGENI.raiseError(res)

    raise POAError(res["output"], res)

  @staticmethod
  def paa (context, url, action, options = None):
    from ..minigcf import amapi3 as AM3

    res = AM3.paa(url, False, context.cf.cert, context.cf.key, action, options)
    if res["code"]["geni_code"] == 0:
      return res["value"]

    raise POAError(res["output"], res)

  @staticmethod
  def allocate (context, url, sname, rspec, options = None):
    if not options: options = {}
    from ..minigcf import amapi3 as AM3

    sinfo = context.getSliceInfo(sname)

    res = AM3.allocate(url, False, context.cf.cert, context.cf.key, [sinfo], sinfo.urn, rspec, options)
    if res["code"]["geni_code"] == 0:
      return res
    if res["code"].has_key("am_type"):
      if res["code"]["am_type"] == "protogeni":
        ProtoGENI.raiseError(res)
    raise AllocateError(res["output"], res)

  @staticmethod
  def provision (context, url, sname, urns = None, options = None):
    from ..minigcf import amapi3 as AM3

    if not options: options = {}
    if urns is not None:
      if not isinstance(urns, list): urns = [urns]

    sinfo = context.getSliceInfo(sname)
    if not urns:
      urns = [sinfo.urn]

    res = AM3.provision(url, False, context.cf.cert, context.cf.key, [sinfo], urns, options)
    if res["code"]["geni_code"] == 0:
      return res
    if res["code"].has_key("am_type"):
      if res["code"]["am_type"] == "protogeni":
        ProtoGENI.raiseError(res)
    raise ProvisionError(res["output"], res)

  @staticmethod
  def delete (context, url, sname, urns, options = None):
    from ..minigcf import amapi3 as AM3

    if not options: options = {}
    if not isinstance(urns, list): urns = [urns]

    sinfo = context.getSliceInfo(sname)

    res = AM3.delete(url, False, context.cf.cert, context.cf.key, [sinfo], urns, options)
    if res["code"]["geni_code"] == 0:
      return res
    if res["code"].has_key("am_type"):
      if res["code"]["am_type"] == "protogeni":
        ProtoGENI.raiseError(res)
    raise ProvisionError(res["output"], res)


class AMAPIv2(object):
  @staticmethod
  def listresources (context, url, sname, options = None):
    if not options: options = {}

    from ..minigcf import amapi2 as AM2
    creds = []

    surn = None
    if sname:
      sinfo = context.getSliceInfo(sname)
      surn = sinfo.urn
      creds.append(open(sinfo.path, "rb").read())

    creds.append(open(context.usercred_path, "rb").read())

    res = AM2.listresources(url, False, context.cf.cert, context.cf.key, creds, options, surn)
    if res["code"]["geni_code"] == 0:
      return res
    if res["code"].has_key("am_type"):
      if res["code"]["am_type"] == "protogeni":
        ProtoGENI.raiseError(res)

    raise ListResourcesError(res["output"], res)

  @staticmethod
  def createsliver (context, url, sname, rspec):
    from ..minigcf import amapi2 as AM2

    sinfo = context.getSliceInfo(sname)
    cred_data = open(sinfo.path, "rb").read()

    udata = []
    for user in context._users:
      data = {"urn" : user.urn, "keys" : [open(x, "rb").read() for x in user._keys]}
      udata.append(data)

    res = AM2.createsliver(url, False, context.cf.cert, context.cf.key, [cred_data], sinfo.urn, rspec, udata)
    if res["code"]["geni_code"] == 0:
      return res
    if res["code"].has_key("am_type"):
      if res["code"]["am_type"] == "protogeni":
        ProtoGENI.raiseError(res)
    raise CreateSliverError(res["output"], res)

  @staticmethod
  def sliverstatus (context, url, sname):
    from ..minigcf import amapi2 as AM2

    sinfo = context.getSliceInfo(sname)
    cred_data = open(sinfo.path, "rb").read()
    res = AM2.sliverstatus(url, False, context.cf.cert, context.cf.key, [cred_data], sinfo.urn)
    if res["code"]["geni_code"] == 0:
      return res["value"]
    if res["code"].has_key("am_type"):
      if res["code"]["am_type"] == "protogeni":
        ProtoGENI.raiseError(res)
    raise SliverStatusError(res["output"], res)

  @staticmethod
  def renewsliver (context, url, sname, date):
    from ..minigcf import amapi2 as AM2

    sinfo = context.getSliceInfo(sname)
    cred_data = open(sinfo.path, "rb").read()
    res = AM2.renewsliver(url, False, context.cf.cert, context.cf.key, [cred_data], sinfo.urn, date)
    if res["code"]["geni_code"] == 0:
      return res["value"]
    raise RenewSliverError(res["output"], res)

  @staticmethod
  def deletesliver (context, url, sname):
    from ..minigcf import amapi2 as AM2

    sinfo = context.getSliceInfo(sname)
    cred_data = open(sinfo.path, "rb").read()
    res = AM2.deletesliver(url, False, context.cf.cert, context.cf.key, [cred_data], sinfo.urn)
    if res["code"]["geni_code"] == 0:
      return res["value"]
    raise DeleteSliverError(res["output"], res)

  @staticmethod
  def getversion (context, url):
    from ..minigcf import amapi2 as AM2

    res = AM2.getversion(url, False, context.cf.cert, context.cf.key)
    if res["code"]["geni_code"] == 0:
      return res["value"]
    raise GetVersionError(res["output"], res)


APIRegistry.register("amapiv2", AMAPIv2())
APIRegistry.register("amapiv3", AMAPIv3())




# Copyright (c) 2014-2015  Barnstormer Softworks, Ltd.

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from __future__ import absolute_import

import sys
import inspect

from .core import AM

class IGOF(AM):
  def __init__ (self, name, host, url = None):
    self.host = host
    if url is None:
      url = "https://%s:3626/foam/gapi/2" % (host)
    super(IGOF, self).__init__(name, url, "amapiv2", "foam")


CaseWestern = IGOF("ig-of-cwru", "foam.geni.case.edu")
CENIC = IGOF("ig-of-cenic", "foam.instageni.cenic.net")
Cornell = IGOF("ig-of-cornell", "foam.geni.it.cornell.edu")
Clemson = IGOF("ig-of-clemson", "foam.instageni.clemson.edu")
Dublin = IGOF("ig-of-ohmetrodc", "foam.instageni.metrodatacenter.com")
GATech = IGOF("ig-of-gatech", "foam.instageni.rnoc.gatech.edu")
GPO = IGOF("ig-of-gpo", "foam.instageni.gpolab.bbn.com")
Illinois = IGOF("ig-of-illinois", "foam.instageni.illinois.edu")
Kansas = IGOF("ig-of-kansas", "foam.instageni.ku.gpeni.net")
Kentucky = IGOF("ig-of-kentucky", "foam.lan.sdn.uky.edu")
Kettering = IGOF("ig-of-kettering", "foam.geni.kettering.edu")
LSU = IGOF("ig-of-lsu", "foam.instageni.lsu.edu")
MAX = IGOF("ig-of-max", "foam.instageni.maxgigapop.net")
Missouri = IGOF("ig-of-missouri", "foam.instageni.rnet.missouri.edu")
MOXI = IGOF("ig-of-moxi", "foam.instageni.iu.edu")
Northwestern = IGOF("ig-of-northwestern", "foam.instageni.northwestern.edu")
NPS = IGOF("ig-of-nps", "foam.instageni.nps.edu")
NYSERNet = IGOF("ig-of-nysernet", "foam.instageni.nysernet.org")
NYU = IGOF("ig-of-nyu", "foam.genirack.nyu.edu")
Princeton = IGOF("ig-of-princeton", "foam.instageni.cs.princeton.edu")
Rutgers = IGOF("ig-of-rutgers", "foam.instageni.rutgers.edu")
SOX = IGOF("ig-of-sox", "foam.instageni.sox.net")
Stanford = IGOF("ig-of-stanford", "foam.instageni.stanford.edu")
UCLA = IGOF("ig-of-ucla", "foam.instageni.idre.ucla.edu")
UMKC = IGOF("ig-of-umkc", "foam.instageni.umkc.edu")
Utah = IGOF("ig-of-utah", "foam.utah.geniracks.net")
UtahDDC = IGOF("ig-of-utahddc", "foam.utahddc.geniracks.net")
Wisconsin = IGOF("ig-of-wisconsin", "foam.instageni.wisc.edu")


def aggregates ():
  module = sys.modules[__name__]
  for _,obj in inspect.getmembers(module):
    if isinstance(obj, AM):
      yield obj

def name_to_aggregate ():
  result = {}
  module = sys.modules[__name__]
  for _,obj in inspect.getmembers(module):
    if isinstance(obj, AM):
      result[obj.name] = obj
  return result
